from pathlib import Path

import pytest
from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import FlowCellError
from cg.meta.demultiplex.utils import create_delivery_file_in_flow_cell_directory
from cg.meta.demultiplex.validation import (
    is_file_path_compressed_fastq,
    is_flow_cell_directory_valid,
    is_lane_in_fastq_file_name,
    is_sample_id_in_directory_name,
    is_valid_sample_fastq_file,
)


def test_validate_sample_fastq_with_valid_file():
    # GIVEN a sample fastq file with a lane number and sample id in the parent directory name
    sample_fastq = Path("Sample_123/sample_L0002.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid = is_valid_sample_fastq_file(sample_fastq=sample_fastq, sample_internal_id="sample")

    # THEN it should be valid
    assert is_valid


def test_validate_sample_fastq_without_sample_id_in_parent_directory_name():
    # GIVEN a sample fastq file without a sample id in the parent directory name or file name
    sample_fastq = Path("L0002.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid_fastq = is_valid_sample_fastq_file(
        sample_fastq=sample_fastq, sample_internal_id="sample_id"
    )

    # THEN it should not be valid
    assert not is_valid_fastq


def test_validate_sample_fastq_without_lane_number_in_path():
    # GIVEN a sample fastq file without a lane number
    sample_fastq = Path("Sample_123/sample_id.fastq.gz")

    # WHEN validating the sample fastq file
    is_valid_fastq = is_valid_sample_fastq_file(sample_fastq, sample_internal_id="sample_id")

    # THEN it should not be valid
    assert not is_valid_fastq


def test_validate_sample_fastq_with_invalid_file_extension():
    # GIVEN a sample fastq file without a valid file extension
    sample_fastq = Path("Sample_123/123_L0002.fastq")

    # WHEN validating the sample fastq file
    is_valid_fastq = is_valid_sample_fastq_file(sample_fastq, sample_internal_id="123")

    # THEN it should not be valid
    assert not is_valid_fastq


def test_is_file_path_compressed_fastq_with_valid_file():
    # GIVEN a valid .fastq.gz file
    file_path = Path("sample_L0002.fastq.gz")

    # WHEN checking if the file path is a compressed fastq file
    result = is_file_path_compressed_fastq(file_path)

    # THEN the result should be True
    assert result is True


def test_is_file_path_compressed_fastq_with_invalid_file():
    # GIVEN a file with invalid extension
    file_path = Path("sample_L0002.fastq")

    # WHEN checking if the file path is a compressed fastq file
    result = is_file_path_compressed_fastq(file_path)

    # THEN the result should be False
    assert result is False


def test_is_lane_in_fastq_file_name_with_valid_file():
    # GIVEN a valid file containing lane number
    file_path = Path("sample_L0002.fastq.gz")

    # WHEN checking if the lane number is in the fastq file name
    result = is_lane_in_fastq_file_name(file_path)

    # THEN the result should be True
    assert result is True


def test_is_lane_in_fastq_file_name_with_invalid_file():
    # GIVEN a file without lane number
    file_path = Path("sample.fastq.gz")

    # WHEN checking if the lane number is in the fastq file name
    result = is_lane_in_fastq_file_name(file_path)

    # THEN the result should be False
    assert result is False


def test_is_sample_id_in_directory_name_with_valid_directory():
    # GIVEN a directory containing sample id
    directory = Path("Sample_123")

    # WHEN checking if the sample id is in the directory name
    result = is_sample_id_in_directory_name(directory=directory, sample_internal_id="123")

    # THEN the result should be True
    assert result is True


def test_is_sample_id_in_directory_name_with_invalid_directory():
    # GIVEN a directory without sample id
    directory = Path("sample/123_L0002.fastq.gz")

    # WHEN checking if the sample id is in the directory name
    result = is_sample_id_in_directory_name(directory=directory, sample_internal_id="sample_id")

    # THEN the result should be False
    assert result is False


def test_is_flow_cell_directory_valid_when_directory_exists_and_demultiplexing_complete(
    tmp_path: Path,
):
    # GIVEN a temporary directory as the flow cell directory with a demux complete file
    flow_cell_directory = tmp_path
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # WHEN validating the flow cell directory
    is_flow_cell_directory_valid(flow_cell_directory)

    # THEN no exception is raised


def test_is_flow_cell_directory_valid_when_directory_exists_but_demultiplexing_not_complete(
    tmp_path: Path,
):
    # GIVEN a temporary directory as the flow cell directory without a demux complete file
    flow_cell_directory = tmp_path

    # WHEN validating the flow cell directory
    with pytest.raises(FlowCellError):
        is_flow_cell_directory_valid(flow_cell_directory)


def test_is_flow_cell_directory_valid_when_directory_does_not_exist():
    # GIVEN a directory that doesn't exist
    flow_cell_directory = Path("non_existent_directory")

    # WHEN validating the flow cell directory
    with pytest.raises(FlowCellError):
        is_flow_cell_directory_valid(flow_cell_directory)



from pathlib import Path

import pytest


@pytest.fixture(name="json_stream")
def fixture_json_stream() -> str:
    """Return string with json format"""
    _content = """{"Lorem": {"ipsum": "sit"}}"""
    return _content


@pytest.fixture(name="json_file_path")
def fixture_json_file_path(fixtures_dir: Path) -> Path:
    """Return a file path to example json file"""
    return Path(fixtures_dir, "io", "example_json.json")


@pytest.fixture(name="json_temp_path")
def fixture_json_temp_path(cg_dir: Path) -> Path:
    """Return a temp file path to use when writing json"""
    return Path(cg_dir, "write_json.json")


@pytest.fixture(name="yaml_stream")
def fixture_yaml_stream() -> str:
    """Return string with yaml format"""
    _content = """- Lorem
- ipsum
- sit
- amet
"""
    return _content


@pytest.fixture(name="csv_file_path")
def fixture_csv_file_path(fixtures_dir: Path) -> Path:
    """Return a file path to example CSV file."""
    return Path(fixtures_dir, "io", "example_csv.csv")


@pytest.fixture(name="csv_stream")
def fixture_csv_stream() -> str:
    """Return string with CSV format."""
    return """Lorem,ipsum,sit,amet"""


@pytest.fixture(name="csv_temp_path")
def fixture_csv_temp_path(cg_dir: Path) -> Path:
    """Return a temp file path to use when writing csv."""
    return Path(cg_dir, "write_csv.csv")


@pytest.fixture(name="xml_file_path")
def fixture_xml_file_path(fixtures_dir: Path) -> Path:
    """Return a file path to example XML file."""
    return Path(fixtures_dir, "io", "example_xml.xml")


@pytest.fixture(name="xml_temp_path")
def fixture_xml_temp_path(cg_dir: Path) -> Path:
    """Return a temp file path to use when writing xml."""
    return Path(cg_dir, "write_xml.xml")



import logging
import re
import shutil
from pathlib import Path
from typing import List, Optional


from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD, Sequencers
from cg.exc import FlowCellError
from cg.meta.demultiplex.validation import (
    is_bcl2fastq_demux_folder_structure,
    is_flow_cell_directory_valid,
    is_valid_sample_fastq_file,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData

from cg.utils.files import get_file_in_directory, get_files_matching_pattern


LOG = logging.getLogger(__name__)


def get_lane_from_sample_fastq(sample_fastq_path: Path) -> int:
    """
    Extract the lane number from the sample fastq path.
    Pre-condition:
        - The fastq file name contains the lane number formatted as _L<lane_number>
    """
    pattern = r"_L(\d+)"
    lane_match = re.search(pattern, sample_fastq_path.name)

    if lane_match:
        return int(lane_match.group(1))

    raise ValueError(f"Could not extract lane number from fastq file name {sample_fastq_path.name}")


def get_sample_fastqs_from_flow_cell(
    flow_cell_directory: Path, sample_internal_id: str
) -> Optional[List[Path]]:
    """Retrieve sample FastQs from a flow cell directory."""
    root_pattern = f"{sample_internal_id}_S*_L*_R*_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    unaligned_pattern = f"Unaligned*/Project_*/Sample_{sample_internal_id}*/*{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    for pattern in [root_pattern, unaligned_pattern]:
        sample_fastqs: List[Path] = get_files_matching_pattern(
            directory=flow_cell_directory, pattern=pattern
        )
        valid_sample_fastqs: List[Path] = get_valid_sample_fastqs(
            fastq_paths=sample_fastqs, sample_internal_id=sample_internal_id
        )

        if valid_sample_fastqs:
            return valid_sample_fastqs


def get_valid_sample_fastqs(fastq_paths: List[Path], sample_internal_id: str) -> List[Path]:
    """Return a list of valid fastq files."""
    return [
        fastq
        for fastq in fastq_paths
        if is_valid_sample_fastq_file(sample_fastq=fastq, sample_internal_id=sample_internal_id)
    ]


def get_bcl_converter_name(flow_cell_directory: Path) -> str:
    if is_bcl2fastq_demux_folder_structure(flow_cell_directory=flow_cell_directory):
        return BclConverter.BCL2FASTQ
    return BclConverter.BCLCONVERT


def create_delivery_file_in_flow_cell_directory(flow_cell_directory: Path) -> None:
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).touch()


def get_q30_threshold(sequencer_type: Sequencers) -> int:
    return FLOWCELL_Q30_THRESHOLD[sequencer_type]


def get_sample_sheet_path(
    flow_cell_directory: Path,
    sample_sheet_file_name: str = DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
) -> Path:
    """Return the path to the sample sheet in the flow cell directory."""
    return get_file_in_directory(directory=flow_cell_directory, file_name=sample_sheet_file_name)


def parse_flow_cell_directory_data(
    flow_cell_directory: Path, bcl_converter: str
) -> FlowCellDirectoryData:
    """Return flow cell data from the flow cell directory."""
    is_flow_cell_directory_valid(flow_cell_directory)
    return FlowCellDirectoryData(flow_cell_path=flow_cell_directory, bcl_converter=bcl_converter)


def copy_sample_sheet(
    sample_sheet_source_directory: Path, sample_sheet_destination_directory: Path
) -> None:
    """Copy the sample sheet from the flow-cell-run dir to demultiplex-runs dir for a flow cell."""
    sample_sheet_source: Path = Path(
        sample_sheet_source_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    sample_sheet_destination: Path = Path(
        sample_sheet_destination_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )

    if not sample_sheet_destination.exists():
        LOG.debug(
            f"Copy sample sheet {sample_sheet_source_directory} from flow cell to demuxed result dir {sample_sheet_destination_directory}"
        )
        try:
            shutil.copy(
                sample_sheet_source.as_posix(),
                sample_sheet_destination.as_posix(),
            )
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Could not copy sample sheet from {sample_sheet_source_directory} to {sample_sheet_destination_directory}: {e}"
            )
        return
    LOG.warning(f"Sample sheet already exists: {sample_sheet_destination}, skipping copy.")



from typing import List

from pydantic.v1 import BaseModel, validator


class NextflowSample(BaseModel):
    """Nextflow samplesheet model.

    Attributes:
        sample: sample name, corresponds to case_id
        fastq_r1: list of all fastq read1 files corresponding to sample
        fastq_r2: list of all fastq read2 files corresponding to sample
    """

    sample: str
    fastq_r1: List[str]
    fastq_r2: List[str]

    @validator("fastq_r2")
    def fastq1_fastq2_len_match(cls, value: List[str], values: dict) -> str:
        """Verify that the number of fastq files is the same for R1 and R2."""
        assert len(value) == len(values.get("fastq_r1")) or len(value) == 0
        return "Length of fastq_r1 and fastq_r2 do not match"



"""Rnafusion analysis model."""
from typing import Dict

from cg.models.analysis import AnalysisModel
from cg.models.rnafusion.metrics import RnafusionQCMetrics


class RnafusionAnalysis(AnalysisModel):
    """Rnafusion analysis model.

    Attributes:
        sample_metrics: retrieved QC metrics associated to a sample
    """

    sample_metrics: Dict[str, RnafusionQCMetrics]



import logging
from typing import List

from cg.apps.cgstats.parsers.conversion_stats import (
    ConversionStats,
    SampleConversionResults,
    UnknownBarcode,
)
from pydantic.v1 import BaseModel
from typing_extensions import Literal

LOG = logging.getLogger(__name__)
MIN_CLUSTER_COUNT = 1000000
DEMUX_REPORT_HEADER: List[str] = ["BARCODE", "SAMPLE", "TYPE", "LANE", "READ_COUNT"]


class SampleData(BaseModel):
    read_count: int
    barcode: str
    lane: int
    sample: str = "Unknown"
    type: Literal["sample", "unknown-barcode"]


def get_low_count_samples(samples: List[SampleConversionResults], lane: int) -> List[SampleData]:
    """return all samples with cluster count below a certain threshold"""
    low_samples: List[SampleData] = []
    for sample in samples:
        if sample.pass_filter_cluster_count >= MIN_CLUSTER_COUNT:
            continue
        low_samples.append(
            SampleData(
                read_count=sample.pass_filter_cluster_count,
                barcode=sample.barcode,
                lane=lane,
                sample=sample.sample_id,
                type="sample",
            )
        )
    return low_samples


def get_unknown_barcodes(unknown_barcodes: List[UnknownBarcode], lane: int) -> List[SampleData]:
    LOG.info(f"Adding unknown barcodes for lane {lane}")
    return [
        SampleData(
            read_count=unknown.read_count,
            barcode=unknown.barcode,
            lane=lane,
            type="unknown-barcode",
        )
        for unknown in unknown_barcodes
    ]


def get_demux_report_data(
    conversion_stats: ConversionStats, skip_empty: bool = False
) -> List[SampleData]:
    """create a list of rows with report data from a ConversionStats object

    If skip_empty the report will be empty if there are no samples with low cluster counts
    """
    report_data: List[SampleData] = []
    for lane, barcode_to_result in conversion_stats.lanes_to_barcode.items():
        lane_samples = get_low_count_samples(
            samples=[barcode_to_result[barcode] for barcode in barcode_to_result], lane=lane
        )
        if lane_samples:
            report_data.extend(lane_samples)
    if skip_empty and not report_data:
        return report_data
    for lane, barcode_data in conversion_stats.lanes_to_unknown_barcode.items():
        report_data.extend(get_unknown_barcodes(unknown_barcodes=barcode_data, lane=lane))
    return report_data


def create_demux_report(conversion_stats: ConversionStats, skip_empty: bool = False) -> List[str]:
    """Find the barcodes from each lane with low number of reads and create a report"""
    report_content: List[str] = ["\t".join(DEMUX_REPORT_HEADER)]
    report_data = get_demux_report_data(conversion_stats=conversion_stats, skip_empty=skip_empty)
    for sample_data in report_data:
        report_content.append(
            "\t".join(
                [
                    sample_data.barcode,
                    sample_data.sample,
                    sample_data.type,
                    str(sample_data.lane),
                    str(sample_data.read_count),
                ]
            )
        )
    return report_content



"""Test demultiplex API."""
from pathlib import Path
from typing import List

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.models.cg_config import CGConfig


def test_get_all_demultiplexed_flow_cell_out_dirs(
    demultiplex_context: CGConfig, demultiplexed_flow_cell_finished_working_directory: Path
):
    """Test returning all flow cell directories from the demultiplexing run directory."""
    # GIVEN a demultiplex flow cell finished output directory that exist

    # GIVEN a demultiplex context
    demux_api: DemultiplexingAPI = demultiplex_context.demultiplex_api

    # WHEN calling get_all_demultiplexed_flow_cell_dirs
    demultiplex_flow_cell_dirs: List[Path] = demux_api.get_all_demultiplexed_flow_cell_dirs()

    # THEN the demultiplexed flow cells run directories should be returned
    assert demultiplex_flow_cell_dirs[0] == demultiplexed_flow_cell_finished_working_directory



"""Module for Tower Analysis API."""

import logging
from pathlib import Path
from typing import Iterable, List

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.utils.utils import build_command_from_dict

LOG = logging.getLogger(__name__)


class TowerAnalysisAPI:
    """Handles communication between tower processes
    and the rest of CG infrastructure."""

    @classmethod
    def get_tower_launch_parameters(cls, tower_pipeline: str, command_args: dict) -> List[str]:
        """Returns a tower launch command given a dictionary with arguments."""

        tower_options: List[str] = build_command_from_dict(
            options={
                f"--{arg.replace('_', '-')}": command_args.get(arg, None)
                for arg in (
                    "work_dir",
                    "profile",
                    "params_file",
                    "config",
                    "name",
                    "revision",
                    "compute_env",
                )
            },
            exclude_true=True,
        )
        return ["launch"] + tower_options + [tower_pipeline]

    @classmethod
    def get_tower_relaunch_parameters(cls, from_tower_id: int, command_args: dict) -> List[str]:
        """Returns a tower relaunch command given a dictionary with arguments."""

        tower_options: List[str] = build_command_from_dict(
            options={
                f"--{arg.replace('_', '-')}": command_args.get(arg, None)
                for arg in (
                    "profile",
                    "params_file",
                    "config",
                    "compute_env",
                )
            },
            exclude_true=True,
        )
        return [
            "runs",
            "relaunch",
            "--id",
            str(from_tower_id),
            "--name",
            f"{command_args.get('name')}_from{str(from_tower_id)}",
        ] + tower_options

    @staticmethod
    def get_tower_id(stdout_lines: Iterable) -> str:
        """Parse the stdout and return a workflow id. An example of the output to parse is:
        Case <CASE_ID> exists in status db
        Running RNAFUSION analysis for <CASE_ID>
        Pipeline will be executed using tower
        Running command <COMMAND>

          Workflow 1uxZE9JM7Tl58r submitted at [<WORKSPACE>] workspace.

        https://<URL_TO_TOWER_CASE>
        Action running set for case <CASE_ID>"""
        for line in stdout_lines:
            if line.strip().startswith("Workflow"):
                return line.strip().split()[1]

    @classmethod
    def get_last_tower_id(cls, case_id: str, trailblazer_config: Path) -> int:
        """Return the previously-stored NF-Tower ID."""
        if not trailblazer_config.exists():
            raise FileNotFoundError(f"No NF-Tower ID found for case {case_id}.")
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=trailblazer_config
        ).get(case_id)[-1]



import logging
from pathlib import Path
from typing import Dict, List, Optional

import alchy
from sqlalchemy import or_

from cg.apps.cgstats.db.models import (
    Supportparams,
    Datasource,
    Demux,
    DemuxSample,
    DragenDemuxSample,
    Flowcell,
    Project,
    Sample,
    Unaligned,
)
from cg.apps.cgstats.parsers.conversion_stats import ConversionStats, SampleConversionResults
from cg.apps.cgstats.parsers.demux_stats import DemuxStats, SampleBarcodeStats
from cg.apps.demultiplex.sample_sheet.models import FlowCellSample, SampleSheet
from cg.models.cgstats.stats_sample import StatsSample
from cg.models.demultiplex.demux_results import DemuxResults

LOG = logging.getLogger(__name__)
SAMPLE_PATTERN = "{}\_%"


class FindHandler:
    def get_support_parameters_by_document_path(
        self, document_path: str
    ) -> Optional[Supportparams]:
        """Get support parameters by document path."""

        LOG.debug(f"Searching for support parameters with file {document_path}")

        support_params: Supportparams = Supportparams.query.filter_by(
            document_path=document_path
        ).first()

        if support_params:
            LOG.debug(f"Found support parameters with id {support_params.supportparams_id}")
        else:
            LOG.debug("Support parameters not found")

        return support_params

    def get_datasource_by_document_path(self, document_path: str) -> Optional[Datasource]:
        """Get data source by document path."""
        LOG.debug(f"Search for datasource with file {document_path}")

        datasource: Datasource = Datasource.query.filter_by(document_path=document_path).first()

        if datasource:
            LOG.debug(f"Found datasource with id {datasource.datasource_id}")
        else:
            LOG.debug("Could not find datasource")

        return datasource

    def get_flow_cell_by_name(self, flow_cell_name: str) -> Optional[Flowcell]:
        """Get flow cell by name."""
        LOG.debug(f"Searching for flow cell {flow_cell_name}")

        flowcell: Flowcell = Flowcell.query.filter_by(flowcellname=flow_cell_name).first()

        if flowcell:
            LOG.debug(f"Found flow cell with id {flowcell.flowcell_id}")
        else:
            LOG.debug("Flow cell not found")

        return flowcell

    def get_demux_by_flow_cell_id_and_base_mask(
        self, flowcell_id: int, base_mask: str = ""
    ) -> Optional[Demux]:
        """Get demux by flow cell id and base mask."""
        return Demux.query.filter_by(flowcell_id=flowcell_id).filter_by(basemask=base_mask).first()

    def get_project_by_name(self, project_name: str) -> Optional[Project]:
        """Get project by name."""
        return Project.query.filter_by(projectname=project_name).first()

    def get_sample(self, sample_id: str):
        """Get a unique demux sample."""
        pattern = SAMPLE_PATTERN.format(sample_id)
        return Sample.query.filter(
            or_(Sample.samplename.like(pattern), Sample.samplename == sample_id)
        ).first()

    def get_sample_by_name_and_barcode(self, sample_name: str, barcode: str) -> Optional[Sample]:
        """Get sample by name and barcode."""
        return Sample.query.filter_by(samplename=sample_name).filter_by(barcode=barcode).first()

    def get_unaligned_by_sample_id_demux_id_and_lane(
        self, sample_id: int, demux_id: int, lane: int
    ) -> Optional[Unaligned]:
        """Get unaligned by sample id, demux id and lane."""
        return (
            Unaligned.query.filter_by(sample_id=sample_id)
            .filter_by(demux_id=demux_id)
            .filter_by(lane=lane)
            .first()
        )

    def get_samples(self, flowcell: str, project_name: Optional[str] = None) -> alchy.Query:
        query: alchy.Query = Sample.query.join(Sample.unaligned, Unaligned.demux, Demux.flowcell)
        if project_name:
            query = query.join(Sample.project).filter(Project.projectname == project_name)

        query = query.filter(Flowcell.flowcellname == flowcell)

        return query

    def project_sample_stats(
        self, flowcell: str, project_name: Optional[str] = None
    ) -> List[StatsSample]:
        samples_query: alchy.Query = self.get_samples(flowcell=flowcell, project_name=project_name)
        db_sample: Sample
        return [StatsSample.from_orm(db_sample) for db_sample in samples_query]

    def get_dragen_demux_samples(
        self,
        demux_results: DemuxResults,
        sample_sheet: SampleSheet,
    ) -> Dict[int, Dict[str, DragenDemuxSample]]:
        """Gather information from Dragen demultiplexing results and create samples with the correct
        information"""
        LOG.info("Gather post Dragen demultiplexing statistics for demultiplexed samples")
        demux_samples: Dict[int, Dict[str, DragenDemuxSample]] = {}

        for sample in sample_sheet.samples:
            if sample.lane not in demux_samples:
                demux_samples[sample.lane] = {}

            demultiplexing_stats = demux_results.demultiplexing_stats.parsed_stats[sample.lane][
                sample.sample_id
            ]

            quality_metrics = demux_results.quality_metrics.parsed_metrics[sample.lane][
                sample.sample_id
            ]

            adapter_metrics = demux_results.adapter_metrics.parsed_metrics[sample.lane][
                sample.sample_id
            ]

            demux_samples[sample.lane][sample.sample_id] = DragenDemuxSample(
                sample_name=sample.sample_id,
                flowcell=demux_results.flow_cell.id,
                lane=sample.lane,
                reads=int(demultiplexing_stats["# Reads"]),
                perfect_reads=int(demultiplexing_stats["# Perfect Index Reads"]),
                one_mismatch_reads=int(demultiplexing_stats["# One Mismatch Index Reads"]),
                pass_filter_q30=int(quality_metrics["YieldQ30"]),
                mean_quality_score=float(quality_metrics["Mean Quality Score (PF)"]),
                r1_sample_bases=int(adapter_metrics["R1_SampleBases"]),
                r2_sample_bases=int(adapter_metrics["R2_SampleBases"]),
                read_length=demux_results.run_info.mean_read_length(),
            )

        return demux_samples

    def get_demux_samples(
        self, conversion_stats: ConversionStats, demux_stats_path: Path, sample_sheet: SampleSheet
    ) -> Dict[int, Dict[str, DemuxSample]]:
        """Gather information from demultiplexing results and create samples with the correct
        information"""
        LOG.info("Gather post demultiplexing statistics for demultiplexed samples")
        demux_samples: Dict[int, Dict[str, DemuxSample]] = {}
        demux_stats: DemuxStats = DemuxStats(demux_stats_path=demux_stats_path)
        raw_clusters: Dict[int, int] = conversion_stats.raw_clusters_per_lane
        flowcell_id: str = conversion_stats.flowcell_id
        sample: FlowCellSample
        for sample in sample_sheet.samples:
            lane: int = sample.lane
            if lane not in demux_samples:
                demux_samples[lane] = {}
            barcode = "+".join([sample.index, sample.index2]) if sample.index2 else sample.index
            sample_barcode_stats: SampleBarcodeStats = demux_stats.lanes_to_barcode[lane][barcode]
            sample_conversion_stats: SampleConversionResults = conversion_stats.lanes_to_barcode[
                lane
            ][barcode]
            lane_raw_cluster: int = raw_clusters[lane]
            sample_id: str = sample.sample_id
            demux_samples[lane][sample_id] = DemuxSample(
                sample_name=sample.sample_id,
                flowcell=flowcell_id,
                lane=lane,
                nr_raw_clusters_=lane_raw_cluster,
                barcode_stats_=sample_barcode_stats,
                conversion_stats_=sample_conversion_stats,
            )

        return demux_samples



import json

import pytest
from pathlib import Path
from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import (
    Bcl2FastqSampleLaneMetrics,
)
from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq import (
    parse_bcl2fastq_sequencing_metrics,
)
from cg.constants.demultiplexing import (
    BCL2FASTQ_METRICS_DIRECTORY_NAME,
    BCL2FASTQ_METRICS_FILE_NAME,
)


def test_parse_valid_bcl2fastq_sequencing_metrics(tmp_path, valid_bcl2fastq_metrics_data):
    """Test that valid stats.json files as generated by bcl2fastq in the expected directory structure can be parsed correctly."""
    # GIVEN a valid directory structure with a valid stats.json file
    valid_dir = tmp_path / "l1t1" / BCL2FASTQ_METRICS_DIRECTORY_NAME
    valid_dir.mkdir(parents=True)
    stats_json_path = valid_dir / BCL2FASTQ_METRICS_FILE_NAME
    stats_json_path.write_text(json.dumps(valid_bcl2fastq_metrics_data))

    # WHEN parsing the directory containing the valid stats.json file
    result = parse_bcl2fastq_sequencing_metrics(flow_cell_dir=tmp_path)

    # THEN a list of Bcl2FastqTileSequencingMetrics models is returned
    assert isinstance(result, list)
    assert all(isinstance(item, Bcl2FastqSampleLaneMetrics) for item in result)

    # THEN the Bcl2FastqTileSequencingMetrics model contains the expected data
    assert result[0].flow_cell_name == valid_bcl2fastq_metrics_data["Flowcell"]


def test_parse_not_found_bcl2fastq_sequencing_metrics():
    """Test that an error is raised when the expected directory structure is not found."""
    # GIVEN a directory structure that does not contain the expected directory structure
    # WHEN parsing the directory containing the valid stats.json file
    # THEN a FileNotFoundError is raised
    with pytest.raises(FileNotFoundError):
        parse_bcl2fastq_sequencing_metrics(flow_cell_dir=Path("/does/not/exist"))



"""CLI for deleting records in statusDB """

import logging

import click
from cg.cli.delete.case import delete_case
from cg.cli.delete.cases import delete_cases
from cg.cli.delete.observations import delete_observations, delete_available_observations

LOG = logging.getLogger(__name__)


@click.group()
def delete():
    """delete database records in CG."""
    LOG.info("Running CG delete")


delete.add_command(delete_case)
delete.add_command(delete_cases)
delete.add_command(delete_observations)
delete.add_command(delete_available_observations)



"""Report field validation helper"""

from typing import Optional

from cg.constants import NA_FIELD
from cg.constants.constants import SCALE_TO_MILLION_READ_PAIRS
from cg.models.report.report import ReportModel


def get_million_read_pairs(reads: int) -> Optional[float]:
    """Represents the number of sequencing reads as millions of read pairs."""
    return (
        round(reads / SCALE_TO_MILLION_READ_PAIRS, 1) if reads or isinstance(reads, int) else None
    )


def get_missing_fields(empty_fields: list, required_fields: list) -> list:
    """Extracts the missing fields that are required to generate successfully the delivery report."""
    missing_fields = list()
    for field in empty_fields:
        if field in required_fields:
            missing_fields.append(field)
    return missing_fields


def get_empty_fields(report_data: dict) -> list:
    """Returns a list of empty report fields."""
    empty_fields = list()
    for field, value in report_data.items():
        if not value or value == NA_FIELD:
            if isinstance(value, bool):
                continue
            empty_fields.append(field)
    return empty_fields


def get_empty_report_data(report_data: ReportModel) -> dict:
    """Retrieve empty fields from a report data model."""
    empty_fields = {
        "report": get_empty_fields(report_data=report_data.dict()),
        "customer": get_empty_fields(report_data=report_data.customer.dict()),
        "case": get_empty_fields(report_data=report_data.case.dict()),
        "applications": {
            app.tag: get_empty_fields(report_data=app.dict())
            for app in report_data.case.applications
            if get_empty_fields(report_data=app.dict())
        },
        "data_analysis": get_empty_fields(report_data=report_data.case.data_analysis.dict()),
        "samples": {
            sample.id: get_empty_fields(report_data=sample.dict())
            for sample in report_data.case.samples
            if get_empty_fields(report_data=sample.dict())
        },
        "methods": {
            sample.id: get_empty_fields(report_data=sample.methods.dict())
            for sample in report_data.case.samples
            if get_empty_fields(report_data=sample.methods.dict())
        },
        "timestamps": {
            sample.id: get_empty_fields(report_data=sample.timestamps.dict())
            for sample in report_data.case.samples
            if get_empty_fields(report_data=sample.timestamps.dict())
        },
        "metadata": {
            sample.id: get_empty_fields(report_data=sample.metadata.dict())
            for sample in report_data.case.samples
            if get_empty_fields(report_data=sample.metadata.dict())
        },
    }
    # Clear empty values
    empty_fields = {k: v for k, v in empty_fields.items() if v}
    return empty_fields


def get_missing_report_data(empty_fields: dict, required_fields: dict) -> dict:
    """Retrieve the missing required fields from a report data model."""
    nested_sources = ["applications", "samples", "methods", "timestamps", "metadata"]
    missing_fields = dict()
    for source in empty_fields:
        if source in nested_sources:
            # Associates application/sample tags/ids to missing fields
            missing_data = {
                tag: get_missing_fields(
                    empty_fields=empty_fields[source][tag],
                    required_fields=required_fields[source][tag],
                )
                for tag in empty_fields[source]
                if get_missing_fields(
                    empty_fields=empty_fields[source][tag],
                    required_fields=required_fields[source][tag],
                )
            }
        else:
            missing_data = get_missing_fields(
                empty_fields=empty_fields[source], required_fields=required_fields[source]
            )
        if missing_data:
            missing_fields.update({source: missing_data})
    return missing_fields



from cg.constants.sequencing import SequencingPlatform
from cg.models.nextflow.sample import NextflowSample


class TaxprofilerSample(NextflowSample):
    """Taxprofiler sample model is used when building the sample sheet.

    Attributes:
        instrument_platform: taxprofiler config file attributes model
    """

    instrument_platform: SequencingPlatform



"""Post-processing Demultiiplex API."""
import logging
import shutil
from contextlib import redirect_stdout
from pathlib import Path
from typing import Iterable, List, Optional

from housekeeper.store.models import Version
from cg.store.models import Sample

from cg.apps.cgstats.crud import create
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.demux_report import create_demux_report
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.sequencing_metrics_parser.api import (
    create_sample_lane_sequencing_metrics_for_flow_cell,
)
from cg.constants.cgstats import STATS_HEADER
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.constants.sequencing import Sequencers
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import FlowCellError
from cg.meta.demultiplex import files
from cg.meta.demultiplex.utils import (
    create_delivery_file_in_flow_cell_directory,
    get_bcl_converter_name,
    get_lane_from_sample_fastq,
    get_q30_threshold,
    get_sample_fastqs_from_flow_cell,
    get_sample_sheet_path,
    parse_flow_cell_directory_data,
    copy_sample_sheet,
)
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_sample_internal_ids_from_sample_sheet,
)
from cg.meta.transfer import TransferFlowCell
from cg.models.cg_config import CGConfig
from cg.models.cgstats.stats_sample import StatsSample
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.store import Store
from cg.store.models import Flowcell, SampleLaneSequencingMetrics
from cg.utils import Process
from cg.utils.files import get_files_matching_pattern


LOG = logging.getLogger(__name__)


class DemuxPostProcessingAPI:
    """Post demultiplexing API class."""

    def __init__(self, config: CGConfig) -> None:
        self.stats_api: StatsAPI = config.cg_stats_api
        self.demux_api: DemultiplexingAPI = config.demultiplex_api
        self.status_db: Store = config.status_db
        self.hk_api: HousekeeperAPI = config.housekeeper_api
        self.transfer_flow_cell_api: TransferFlowCell = TransferFlowCell(
            db=self.status_db, stats_api=self.stats_api, hk_api=self.hk_api
        )
        self.dry_run = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"Set dry run to {dry_run}")
        self.dry_run = dry_run
        if dry_run:
            self.demux_api.set_dry_run(dry_run=dry_run)

    def transfer_flow_cell(
        self, flow_cell_dir: Path, flow_cell_id: str, store: bool = True
    ) -> None:
        """Transfer flow cell and sequencing files."""
        flow_cell: Flowcell = self.transfer_flow_cell_api.transfer(
            flow_cell_dir=flow_cell_dir, flow_cell_id=flow_cell_id, store=store
        )
        if self.dry_run:
            LOG.info("Dry run will not commit flow cell to database")
            return
        self.status_db.session.add(flow_cell)
        self.status_db.session.commit()

        LOG.info(f"Flow cell added: {flow_cell}")

    def finish_flow_cell_temp(self, flow_cell_directory_name: str) -> None:
        """Store data for the demultiplexed flow cell and mark it as ready for delivery.

        This function:
            - Copies the sample sheet to the demultiplexed flow cell directory
            - Parses and validates the flow cell directory data
            - Stores the flow cell in the status database
            - Stores sequencing metrics in the status database
            - Updates sample read counts in the status database
            - Stores the flow cell data in the housekeeper database
            - Creates a delivery file in the flow cell directory

        Args:
            flow_cell_directory_name (str): The name of the flow cell directory to be finalized.

        Raises:
            FlowCellError: If the flow cell directory or the data it contains is not valid.
        """
        if self.dry_run:
            LOG.info(f"Dry run will not finish flow cell {flow_cell_directory_name}")
            return
        LOG.info(f"Finish flow cell {flow_cell_directory_name}")

        flow_cell_directory: Path = Path(self.demux_api.out_dir, flow_cell_directory_name)
        flow_cell_run_directory: Path = Path(self.demux_api.run_dir, flow_cell_directory_name)

        bcl_converter: str = get_bcl_converter_name(flow_cell_directory)

        parsed_flow_cell: FlowCellDirectoryData = parse_flow_cell_directory_data(
            flow_cell_directory=flow_cell_directory,
            bcl_converter=bcl_converter,
        )

        copy_sample_sheet(
            sample_sheet_source_directory=flow_cell_run_directory,
            sample_sheet_destination_directory=flow_cell_directory,
        )

        try:
            self.store_flow_cell_data(parsed_flow_cell)
        except Exception as e:
            LOG.error(f"Failed to store flow cell data: {str(e)}")
            raise

        create_delivery_file_in_flow_cell_directory(flow_cell_directory)

    def finish_all_flow_cells_temp(self) -> None:
        """Finish all flow cells that need it."""
        flow_cell_dirs = self.demux_api.get_all_demultiplexed_flow_cell_dirs()
        for flow_cell_dir in flow_cell_dirs:
            try:
                self.finish_flow_cell_temp(flow_cell_dir.name)
            except FlowCellError as e:
                LOG.error(f"Failed to finish flow cell {flow_cell_dir.name}: {str(e)}")
                continue

    def store_flow_cell_data(self, parsed_flow_cell: FlowCellDirectoryData) -> None:
        """Store data from the flow cell directory in status db and housekeeper."""
        self.store_flow_cell_data_in_status_db(parsed_flow_cell)
        self.store_sequencing_metrics_in_status_db(parsed_flow_cell)
        self.update_sample_read_counts_in_status_db(parsed_flow_cell)
        self.store_flow_cell_data_in_housekeeper(parsed_flow_cell)

    def store_sequencing_metrics_in_status_db(self, flow_cell: FlowCellDirectoryData) -> None:
        sample_lane_sequencing_metrics: List[
            SampleLaneSequencingMetrics
        ] = create_sample_lane_sequencing_metrics_for_flow_cell(
            flow_cell_directory=flow_cell.path,
            bcl_converter=flow_cell.bcl_converter,
        )
        self.add_sequencing_metrics_to_statusdb(sample_lane_sequencing_metrics)

        LOG.info(f"Added sequencing metrics to status db for: {flow_cell.id}")

    def add_sequencing_metrics_to_statusdb(
        self, sample_lane_sequencing_metrics: List[SampleLaneSequencingMetrics]
    ) -> None:
        for metric in sample_lane_sequencing_metrics:
            if not self.metric_exists_in_status_db(metric):
                self.add_metric_to_status_db(metric)
        self.status_db.session.commit()

    def metric_exists_in_status_db(self, metric: SampleLaneSequencingMetrics) -> bool:
        existing_metrics_entry: Optional[
            SampleLaneSequencingMetrics
        ] = self.status_db.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
            flow_cell_name=metric.flow_cell_name,
            sample_internal_id=metric.sample_internal_id,
            lane=metric.flow_cell_lane_number,
        )
        if existing_metrics_entry:
            LOG.warning(
                f"Sample lane sequencing metrics already exist for {metric.flow_cell_name}, {metric.sample_internal_id}, and {metric.flow_cell_lane_number}. Skipping."
            )
        return bool(existing_metrics_entry)

    def add_metric_to_status_db(self, metric: SampleLaneSequencingMetrics) -> None:
        LOG.debug(
            f"Adding sample lane sequencing metrics for {metric.flow_cell_name}, {metric.sample_internal_id}, and {metric.flow_cell_lane_number}."
        )
        self.status_db.session.add(metric)

    def update_sample_read_counts_in_status_db(self, flow_cell_data: FlowCellDirectoryData) -> None:
        """Update samples in status db with the sum of all read counts for the sample in the sequencing metrics table."""

        q30_threshold: int = get_q30_threshold(flow_cell_data.sequencer_type)
        sample_internal_ids: List[str] = get_sample_internal_ids_from_sample_sheet(
            sample_sheet_path=flow_cell_data.sample_sheet_path,
            flow_cell_sample_type=flow_cell_data.sample_type,
        )

        for sample_id in sample_internal_ids:
            self.update_sample_read_count(sample_id=sample_id, q30_threshold=q30_threshold)

        self.status_db.session.commit()

    def update_sample_read_count(self, sample_id: str, q30_threshold: int) -> None:
        """Update the read count for a sample in status db with all reads exceeding the q30 threshold from the sequencing metrics table."""
        sample: Optional[Sample] = self.status_db.get_sample_by_internal_id(sample_id)

        if sample:
            sample_read_count: int = (
                self.status_db.get_number_of_reads_for_sample_passing_q30_threshold(
                    sample_internal_id=sample_id,
                    q30_threshold=q30_threshold,
                )
            )
            LOG.debug(f"Updating sample {sample_id} with read count {sample_read_count}")
            sample.calculated_read_count = sample_read_count
        else:
            LOG.warning(f"Cannot find {sample_id} in status_db when adding read counts. Skipping.")

    def store_flow_cell_data_in_housekeeper(self, flow_cell: FlowCellDirectoryData) -> None:
        LOG.info(f"Add flow cell data to Housekeeper for {flow_cell.id}")

        self.add_bundle_and_version_if_non_existent(bundle_name=flow_cell.id)

        tags: List[str] = [SequencingFileTag.FASTQ, SequencingFileTag.SAMPLE_SHEET, flow_cell.id]
        self.add_tags_if_non_existent(tag_names=tags)

        self.add_sample_sheet_path_to_housekeeper(
            flow_cell_directory=flow_cell.path, flow_cell_name=flow_cell.id
        )
        self.add_sample_fastq_files_to_housekeeper(flow_cell)
        self.add_demux_logs_to_housekeeper(flow_cell)

    def add_demux_logs_to_housekeeper(self, flow_cell: FlowCellDirectoryData) -> None:
        """Add demux logs to Housekeeper."""
        log_pattern: str = r"*_demultiplex.std*"
        demux_log_file_paths: List[Path] = get_files_matching_pattern(
            directory=Path(self.demux_api.run_dir, flow_cell.full_name), pattern=log_pattern
        )

        tag_names: List[str] = [SequencingFileTag.DEMUX_LOG, flow_cell.id]
        for file_path in demux_log_file_paths:
            try:
                self.add_file_to_bundle_if_non_existent(
                    file_path=file_path, bundle_name=flow_cell.id, tag_names=tag_names
                )
                LOG.info(f"Added demux log file {file_path} to Housekeeper.")
            except FileNotFoundError as e:
                LOG.error(f"Cannot find demux log file {file_path}. Error: {e}.")

    def add_sample_fastq_files_to_housekeeper(self, flow_cell: FlowCellDirectoryData) -> None:
        """Add sample fastq files from flow cell to Housekeeper."""

        sample_internal_ids: List[str] = get_sample_internal_ids_from_sample_sheet(
            sample_sheet_path=flow_cell.sample_sheet_path,
            flow_cell_sample_type=flow_cell.sample_type,
        )

        for sample_internal_id in sample_internal_ids:
            self.add_bundle_and_version_if_non_existent(sample_internal_id)

            sample_fastq_paths: Optional[List[Path]] = get_sample_fastqs_from_flow_cell(
                flow_cell_directory=flow_cell.path, sample_internal_id=sample_internal_id
            )

            if not sample_fastq_paths:
                LOG.warning(
                    f"Cannot find fastq files for sample {sample_internal_id} in {flow_cell.path}. Skipping."
                )
                continue

            for sample_fastq_path in sample_fastq_paths:
                self.store_fastq_path_in_housekeeper(
                    sample_internal_id=sample_internal_id,
                    sample_fastq_path=sample_fastq_path,
                    flow_cell=flow_cell,
                )

    def store_fastq_path_in_housekeeper(
        self, sample_internal_id: str, sample_fastq_path: Path, flow_cell: FlowCellDirectoryData
    ) -> None:
        sample_fastq_should_be_stored: bool = (
            self.check_if_fastq_path_should_be_stored_in_housekeeper(
                sample_id=sample_internal_id,
                sample_fastq_path=sample_fastq_path,
                sequencer_type=flow_cell.sequencer_type,
                flow_cell_name=flow_cell.id,
            )
        )

        if sample_fastq_should_be_stored:
            self.add_file_to_bundle_if_non_existent(
                file_path=sample_fastq_path,
                bundle_name=sample_internal_id,
                tag_names=[SequencingFileTag.FASTQ, flow_cell.id],
            )

    def check_if_fastq_path_should_be_stored_in_housekeeper(
        self,
        sample_id: str,
        sample_fastq_path: Path,
        sequencer_type: Sequencers,
        flow_cell_name: str,
    ) -> bool:
        """
        Check if a sample fastq file should be tracked in Housekeeper.
        Only fastq files that pass the q30 threshold should be tracked.
        """
        lane = get_lane_from_sample_fastq(sample_fastq_path)
        q30_threshold: int = get_q30_threshold(sequencer_type)

        metric = self.status_db.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
            flow_cell_name=flow_cell_name,
            sample_internal_id=sample_id,
            lane=lane,
        )

        if metric:
            return metric.sample_base_fraction_passing_q30 >= q30_threshold / 100

        LOG.warning(
            f"Skipping fastq file {sample_fastq_path.name} as no metrics entry was found in status db."
        )
        LOG.warning(f"Flow cell name: {flow_cell_name}, sample id: {sample_id}, lane: {lane} ")
        return False

    def add_sample_sheet_path_to_housekeeper(
        self, flow_cell_directory: Path, flow_cell_name: str
    ) -> None:
        """Add sample sheet path to Housekeeper."""

        try:
            sample_sheet_file_path: Path = get_sample_sheet_path(
                flow_cell_directory=flow_cell_directory
            )

            self.add_file_to_bundle_if_non_existent(
                file_path=sample_sheet_file_path,
                bundle_name=flow_cell_name,
                tag_names=[SequencingFileTag.SAMPLE_SHEET, flow_cell_name],
            )
        except FileNotFoundError as e:
            LOG.error(
                f"Sample sheet for flow cell {flow_cell_name} in {flow_cell_directory} was not found, error: {e}"
            )

    def add_bundle_and_version_if_non_existent(self, bundle_name: str) -> None:
        """Add bundle if it does not exist."""
        if not self.hk_api.bundle(name=bundle_name):
            self.hk_api.create_new_bundle_and_version(name=bundle_name)

    def add_tags_if_non_existent(self, tag_names: List[str]) -> None:
        """Ensure that tags exist in Housekeeper."""
        for tag_name in tag_names:
            if self.hk_api.get_tag(name=tag_name) is None:
                self.hk_api.add_tag(name=tag_name)

    def add_file_to_bundle_if_non_existent(
        self, file_path: Path, bundle_name: str, tag_names: List[str]
    ) -> None:
        """Add file to Housekeeper if it has not already been added."""
        if not file_path.exists():
            LOG.warning(f"File does not exist: {file_path}")
            return

        if not self.file_exists_in_latest_version_for_bundle(
            file_path=file_path, bundle_name=bundle_name
        ):
            self.hk_api.add_and_include_file_to_latest_version(
                bundle_name=bundle_name,
                file=file_path,
                tags=tag_names,
            )

    def file_exists_in_latest_version_for_bundle(self, file_path: Path, bundle_name: str) -> bool:
        """Check if file exists in latest version for bundle."""
        latest_version: Version = self.hk_api.get_latest_bundle_version(bundle_name=bundle_name)
        return any(
            file_path.name == Path(bundle_file.path).name for bundle_file in latest_version.files
        )

    def store_flow_cell_data_in_status_db(self, parsed_flow_cell: FlowCellDirectoryData) -> None:
        """Create flow cell from the parsed and validated flow cell data."""
        if not self.status_db.get_flow_cell_by_name(flow_cell_name=parsed_flow_cell.id):
            flow_cell: Flowcell = Flowcell(
                name=parsed_flow_cell.id,
                sequencer_type=parsed_flow_cell.sequencer_type,
                sequencer_name=parsed_flow_cell.machine_name,
                sequenced_at=parsed_flow_cell.run_date,
            )
            self.status_db.session.add(flow_cell)
            self.status_db.session.commit()
            LOG.info(f"Flow cell added to status db: {parsed_flow_cell.id}.")
        else:
            LOG.info(f"Flow cell already exists in status db: {parsed_flow_cell.id}. Skipping.")


class DemuxPostProcessingHiseqXAPI(DemuxPostProcessingAPI):
    """Post demultiplexing API class for Hiseq X flow cell."""

    def add_to_cgstats(self, flow_cell_path: Path) -> None:
        """Add flow cell to cgstats."""
        LOG.info(
            f"{self.stats_api.binary} --database {self.stats_api.db_uri} add --machine X -u Unaligned {flow_cell_path.as_posix()}"
        )
        if self.dry_run:
            LOG.info("Dry run will not add flow cell stats")
            return
        cgstats_add_parameters = [
            "--database",
            self.stats_api.db_uri,
            "add",
            "--machine",
            "X",
            "-u",
            "Unaligned",
            flow_cell_path.as_posix(),
        ]
        cgstats_process: Process = Process(binary=self.stats_api.binary)
        cgstats_process.run_command(parameters=cgstats_add_parameters, dry_run=self.dry_run)

    def cgstats_select_project(self, flow_cell_id: str, flow_cell_path: Path) -> None:
        """Process selected project using cgstats."""
        unaligned_dir: Path = Path(flow_cell_path, DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME)
        for project_dir in unaligned_dir.glob("Project_*[0-9]*"):
            (_, project_id) = project_dir.name.split("_")
            stdout_file: Path = Path(
                flow_cell_path, "-".join(["stats", project_id, flow_cell_id]) + ".txt"
            )
            LOG.info(
                f"{self.stats_api.binary} --database {self.stats_api.db_uri} select --project {project_id} {flow_cell_id}"
            )
            if self.dry_run:
                LOG.info("Dry run will not process selected project")
                return
            cgstats_select_parameters: List[str] = [
                "--database",
                self.stats_api.db_uri,
                "select",
                "--project",
                project_id,
                flow_cell_id,
            ]
            cgstats_process: Process = Process(binary=self.stats_api.binary)
            with open(stdout_file.as_posix(), "w") as file:
                with redirect_stdout(file):
                    cgstats_process.run_command(
                        parameters=cgstats_select_parameters, dry_run=self.dry_run
                    )

    def cgstats_lanestats(self, flow_cell_path: Path) -> None:
        """Process lane stats using cgstats."""
        LOG.info(
            f"{self.stats_api.binary} --database {self.stats_api.db_uri} lanestats {flow_cell_path.as_posix()}"
        )
        if self.dry_run:
            LOG.info("Dry run will not add lane stats")
            return
        cgstats_lane_parameters: List[str] = [
            "--database",
            self.stats_api.db_uri,
            "lanestats",
            flow_cell_path.as_posix(),
        ]
        cgstats_process: Process = Process(binary=self.stats_api.binary)
        cgstats_process.run_command(parameters=cgstats_lane_parameters, dry_run=self.dry_run)

    def post_process_flow_cell(self, demux_results: DemuxResults) -> None:
        """Run all the necessary steps for post-processing a demultiplexed flow cell."""
        if not self.dry_run:
            Path(demux_results.flow_cell.path, DemultiplexingDirsAndFiles.DELIVERY).touch()
            # delete cgstats
        self.add_to_cgstats(flow_cell_path=demux_results.flow_cell.path)
        self.cgstats_select_project(
            flow_cell_id=demux_results.flow_cell.id, flow_cell_path=demux_results.flow_cell.path
        )
        self.cgstats_lanestats(flow_cell_path=demux_results.flow_cell.path)
        self.transfer_flow_cell(
            flow_cell_dir=demux_results.flow_cell.path, flow_cell_id=demux_results.flow_cell.id
        )

    def finish_flow_cell(
        self, bcl_converter: str, flow_cell_name: str, flow_cell_path: Path
    ) -> None:
        """Post-processing flow cell."""
        LOG.info(f"Check demultiplexed flow cell {flow_cell_name}")
        try:
            flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
                flow_cell_path=flow_cell_path, bcl_converter=bcl_converter
            )
        except FlowCellError:
            return
        demux_results: DemuxResults = DemuxResults(
            demux_dir=Path(self.demux_api.out_dir, flow_cell_name),
            flow_cell=flow_cell,
            bcl_converter=bcl_converter,
        )
        if not demux_results.flow_cell.is_hiseq_x_copy_completed():
            LOG.info(f"{flow_cell_name} is not yet completely copied")
            return
        if demux_results.flow_cell.is_hiseq_x_delivery_started():
            LOG.info(f"{flow_cell_name} copy is complete and delivery has already started")
            return
        if not demux_results.flow_cell.is_hiseq_x():
            LOG.info(f"{flow_cell_name} is not an Hiseq X flow cell")
            return
        LOG.info(f"{flow_cell_name} copy is complete and delivery will start")
        self.post_process_flow_cell(demux_results=demux_results)

    def finish_all_flow_cells(self, bcl_converter: str) -> None:
        """Loop over all flow cells and post process those that need it."""
        for flow_cell_dir in self.demux_api.get_all_demultiplexed_flow_cell_dirs():
            self.finish_flow_cell(
                bcl_converter=bcl_converter,
                flow_cell_name=flow_cell_dir.name,
                flow_cell_path=flow_cell_dir,
            )


class DemuxPostProcessingNovaseqAPI(DemuxPostProcessingAPI):
    """Post demultiplexing API class for Novaseq flow cell."""

    def rename_files(self, demux_results: DemuxResults) -> None:
        """Rename the files according to how we want to have it after demultiplexing is ready."""
        LOG.info(f"Renaming files for flow cell {demux_results.flow_cell.full_name}")
        flow_cell_id: str = demux_results.flow_cell.id
        for project_dir in demux_results.raw_projects:
            files.rename_project_directory(
                project_directory=project_dir, flow_cell_id=flow_cell_id, dry_run=self.dry_run
            )

    def add_to_cgstats(self, demux_results: DemuxResults) -> None:
        """Add the information from demultiplexing to cgstats"""
        create.create_novaseq_flowcell(manager=self.stats_api, demux_results=demux_results)

    def fetch_report_samples(self, flow_cell_id: str, project_name: str) -> List[StatsSample]:
        samples: List[StatsSample] = self.stats_api.find_handler.project_sample_stats(
            flowcell=flow_cell_id, project_name=project_name
        )
        LOG.info(
            "Found samples: %s for flow cell: %s, project: %s",
            ",".join(sample.sample_name for sample in samples),
            flow_cell_id,
            project_name,
        )
        return samples

    @staticmethod
    def sample_to_report_line(stats_sample: StatsSample, flow_cell_id: str) -> str:
        lanes: str = ",".join(str(lane) for lane in stats_sample.lanes)
        return "\t".join(
            [
                stats_sample.sample_name,
                flow_cell_id,
                lanes,
                ",".join(str(unaligned.read_count) for unaligned in stats_sample.unaligned),
                str(stats_sample.read_count_sum),
                ",".join(str(unaligned.yield_mb) for unaligned in stats_sample.unaligned),
                str(stats_sample.sum_yield),
                ",".join(str(unaligned.q30_bases_pct) for unaligned in stats_sample.unaligned),
                ",".join(str(unaligned.mean_quality_score) for unaligned in stats_sample.unaligned),
            ]
        )

    @staticmethod
    def get_report_lines(stats_samples: List[StatsSample], flow_cell_id: str) -> Iterable[str]:
        """Convert stats samples to format lines ready to print."""
        for stats_sample in sorted(stats_samples, key=lambda x: x.sample_name):
            yield DemuxPostProcessingNovaseqAPI.sample_to_report_line(
                stats_sample=stats_sample, flow_cell_id=flow_cell_id
            )

    @staticmethod
    def write_report(report_path: Path, report_data: List[str]) -> None:
        LOG.info("Write demux report to %s", report_path)
        with report_path.open("w") as report_file:
            report_file.write("\t".join(STATS_HEADER) + "\n")
            for line in report_data:
                report_file.write(line + "\n")

    def get_report_data(self, flow_cell_id: str, project_name: Optional[str] = None) -> List[str]:
        """Fetch the lines that are used to make a report from a flow cell."""
        LOG.info(f"Fetch report data for flow cell: {flow_cell_id} project: {project_name}")
        project_samples: List[StatsSample] = self.fetch_report_samples(
            flow_cell_id=flow_cell_id, project_name=project_name
        )

        return list(self.get_report_lines(stats_samples=project_samples, flow_cell_id=flow_cell_id))

    def create_cgstats_reports(self, demux_results: DemuxResults) -> None:
        """Create a report for every project that was demultiplexed."""
        flow_cell_id: str = demux_results.flow_cell.id
        for project in demux_results.projects:
            project_name: str = project.split("_")[-1]
            report_data: List[str] = self.get_report_data(
                flow_cell_id=flow_cell_id, project_name=project_name
            )
            report_path: Path = demux_results.demux_dir / f"stats-{project_name}-{flow_cell_id}.txt"
            self.write_report(report_path=report_path, report_data=report_data)

    @staticmethod
    def create_barcode_summary_report(demux_results: DemuxResults) -> None:
        report_content = create_demux_report(conversion_stats=demux_results.conversion_stats)
        report_path: Path = demux_results.barcode_report
        if report_path.exists():
            LOG.warning("Report path already exists!")
            return
        LOG.info("Write demux report to %s", report_path)
        with report_path.open("w") as report_file:
            report_file.write("\n".join(report_content))

    @staticmethod
    def copy_sample_sheet(demux_results: DemuxResults) -> None:
        """Copy the sample sheet from run dir to demux dir"""
        LOG.info(
            f"Copy sample sheet {demux_results.sample_sheet_path} from flow cell to demuxed result dir {demux_results.demux_sample_sheet_path}"
        )
        shutil.copy(
            demux_results.sample_sheet_path.as_posix(),
            demux_results.demux_sample_sheet_path.as_posix(),
        )

    def post_process_flow_cell(self, demux_results: DemuxResults) -> None:
        """Run all the necessary steps for post-processing a demultiplexed flow cell.
        This will
            1. Rename all the necessary files and folders
            2. Add the demux results to cgstats
            3. Produce reports for every project
            4. Generate a report with samples that have low cluster count
        """
        if demux_results.files_renamed():
            LOG.info("Files have already been renamed")
        else:
            # not needed
            self.rename_files(demux_results=demux_results)
        # depcrecated
        self.add_to_cgstats(demux_results=demux_results)
        # deprecated
        self.create_cgstats_reports(demux_results=demux_results)
        if demux_results.bcl_converter == "bcl2fastq":
            # relevant - topUnknownbarcodes
            self.create_barcode_summary_report(demux_results=demux_results)
        # moving not needed with new workflow   - on hold
        self.copy_sample_sheet(demux_results=demux_results)

        self.transfer_flow_cell(
            flow_cell_dir=demux_results.flow_cell.path, flow_cell_id=demux_results.flow_cell.id
        )

    def finish_flow_cell(
        self, flow_cell_name: str, bcl_converter: str, force: bool = False
    ) -> None:
        """Go through the post-processing steps for a flow cell.

        Force is used to finish a flow cell even if the files are renamed already.
        """
        LOG.info(
            f"Check demuxed flow cell {flow_cell_name}",
        )
        try:
            flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
                flow_cell_path=Path(self.demux_api.run_dir, flow_cell_name),
                bcl_converter=bcl_converter,
            )
        except FlowCellError:
            LOG.warning(f"Could not find flow cell {flow_cell_name}")
            return
        if not self.demux_api.is_demultiplexing_completed(flow_cell=flow_cell):
            LOG.warning("Demultiplex is not ready for %s", flow_cell_name)
            return

        demux_results: DemuxResults = DemuxResults(
            demux_dir=Path(self.demux_api.out_dir, flow_cell_name),
            flow_cell=flow_cell,
            bcl_converter=bcl_converter,
        )
        if not demux_results.results_dir.exists():
            LOG.warning(f"Could not find results directory {demux_results.results_dir}")
            LOG.info(f"Can not finish flow cell {flow_cell_name}")
            return
        if demux_results.files_renamed():
            LOG.warning("Flow cell is already finished!")
            if not force:
                return

            LOG.info("Post processing flow cell anyway")
        self.post_process_flow_cell(demux_results=demux_results)

    def finish_all_flow_cells(self, bcl_converter: str) -> None:
        """Loop over all flow cells and post-process those that need it."""
        for flow_cell_dir in self.demux_api.get_all_demultiplexed_flow_cell_dirs():
            self.finish_flow_cell(flow_cell_name=flow_cell_dir.name, bcl_converter=bcl_converter)



from pathlib import Path

import pytest


@pytest.fixture(name="correct_flow_cell_name")
def fixture_correct_flow_cell_name(bcl2fastq_flow_cell_full_name: str) -> str:
    """Correct flow cell name."""
    return bcl2fastq_flow_cell_full_name


@pytest.fixture(name="incorrect_flow_cell_too_long")
def fixture_incorrect_flow_cell_too_long(correct_flow_cell_name: str) -> str:
    """Incorrect flow cell name."""
    return correct_flow_cell_name + "_r"


@pytest.fixture(name="incorrect_flow_cell_extension")
def fixture_incorrect_flow_cell_extension(correct_flow_cell_name: str) -> str:
    """Incorrect flow cell name."""
    return correct_flow_cell_name + ".someextension"


@pytest.fixture(name="incorrect_flow_cell_name")
def fixture_incorrect_flow_cell_name() -> str:
    """Incorrect flow cell name."""
    return "201203_A00689_0200_AZZZZZZZZZ"


@pytest.fixture(name="correct_flow_cell_path")
def fixture_correct_flow_cell_path(
    demultiplexed_flow_cells_working_directory, correct_flow_cell_name
) -> Path:
    """Full path to a correctly named flow cell directory in demultiplexed-runs."""
    return Path(demultiplexed_flow_cells_working_directory, correct_flow_cell_name)


@pytest.fixture(name="incorrect_flow_cell_path_too_long")
def fixture_incorrect_flow_cell_path_too_long(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_too_long
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_too_long)


@pytest.fixture(name="incorrect_flow_cell_path_extension")
def fixture_incorrect_flow_cell_path_extension(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_extension
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_extension)


@pytest.fixture(name="non-existent_flow_cell_path")
def fixture_non_existent_flow_cell_path(
    demultiplexed_flow_cells_working_directory, incorrect_flow_cell_name
) -> Path:
    """Full path to an incorrectly named flow cell directory in demultiplexed-runs."""
    return Path(demultiplexed_flow_cells_working_directory, incorrect_flow_cell_name)



"""Fixtures for cli workflow taxprofiler tests"""

import pytest
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig
from pathlib import Path


@pytest.fixture(name="taxprofiler_dir")
def taxprofiler_dir(tmpdir_factory, apps_dir: Path) -> str:
    """Return the path to the taxprofiler apps dir."""
    taxprofiler_dir = tmpdir_factory.mktemp("taxprofiler")
    return Path(taxprofiler_dir).absolute().as_posix()


@pytest.fixture(scope="function", name="taxprofiler_context")
def fixture_taxprofiler_context(
    cg_context: CGConfig,
) -> CGConfig:
    """Context to use in cli."""
    cg_context.meta_apis["analysis_api"] = TaxprofilerAnalysisAPI(config=cg_context)
    return cg_context



import logging
from pathlib import Path
from typing import Dict, List, Type
from pydantic.v1 import parse_obj_as

from cg.apps.demultiplex.sample_sheet.models import FlowCellSample, SampleSheet
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import (
    SampleSheetNovaSeq6000Sections,
    SampleSheetNovaSeqXSections,
)

from cg.exc import SampleSheetError
from cg.io.controller import ReadFile
import re

LOG = logging.getLogger(__name__)


def validate_samples_are_unique(samples: List[FlowCellSample]) -> None:
    """Validate that each sample only exists once."""
    sample_ids: set = set()
    for sample in samples:
        sample_id: str = sample.sample_id.split("_")[0]
        if sample_id in sample_ids:
            message: str = f"Sample {sample.sample_id} exists multiple times in sample sheet"
            LOG.warning(message)
            raise SampleSheetError(message)
        sample_ids.add(sample_id)


def validate_samples_unique_per_lane(samples: List[FlowCellSample]) -> None:
    """Validate that each sample only exists once per lane in a sample sheet."""
    sample_by_lane: Dict[int, List[FlowCellSample]] = get_samples_by_lane(samples)
    for lane, lane_samples in sample_by_lane.items():
        LOG.info(f"Validate that samples are unique in lane {lane}")
        validate_samples_are_unique(samples=lane_samples)


def is_valid_sample_internal_id(sample_internal_id: str) -> bool:
    """
    Check if a sample internal id has the correct structure:
    starts with three letters followed by at least three digits.
    """
    return bool(re.search(r"^[A-Za-z]{3}\d{3}", sample_internal_id))


def get_sample_sheet_from_file(
    infile: Path,
    flow_cell_sample_type: Type[FlowCellSample],
) -> SampleSheet:
    """Parse and validate a sample sheet from file."""
    sample_sheet_content: List[List[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=infile
    )
    return get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=flow_cell_sample_type,
    )


def get_validated_sample_sheet(
    sample_sheet_content: List[List[str]],
    sample_type: Type[FlowCellSample],
) -> SampleSheet:
    """Return a validated sample sheet object."""
    raw_samples: List[Dict[str, str]] = get_raw_samples(sample_sheet_content=sample_sheet_content)
    samples = parse_obj_as(List[sample_type], raw_samples)
    validate_samples_unique_per_lane(samples=samples)
    return SampleSheet(samples=samples)


def get_raw_samples(sample_sheet_content: List[List[str]]) -> List[Dict[str, str]]:
    """Return the samples in a sample sheet as a list of dictionaries."""
    header: List[str] = []
    raw_samples: List[Dict[str, str]] = []

    for line in sample_sheet_content:
        # Skip lines that are too short to contain samples
        if len(line) <= 5:
            continue
        if line[0] in [
            SampleSheetNovaSeq6000Sections.Data.FLOW_CELL_ID.value,
            SampleSheetNovaSeqXSections.Data.LANE.value,
        ]:
            header = line
            continue
        if not header:
            continue
        raw_samples.append(dict(zip(header, line)))
    if not header:
        message = "Could not find header in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    if not raw_samples:
        message = "Could not find any samples in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    return raw_samples


def get_samples_by_lane(
    samples: List[FlowCellSample],
) -> Dict[int, List[FlowCellSample]]:
    """Group and return samples by lane."""
    LOG.debug("Order samples by lane")
    sample_by_lane: Dict[int, List[FlowCellSample]] = {}
    for sample in samples:
        if sample.lane not in sample_by_lane:
            sample_by_lane[sample.lane] = []
        sample_by_lane[sample.lane].append(sample)
    return sample_by_lane


def get_sample_internal_ids_from_sample_sheet(
    sample_sheet_path: Path, flow_cell_sample_type: Type[FlowCellSample]
) -> List[str]:
    """Return the sample internal ids for samples in the sample sheet."""
    sample_sheet = get_sample_sheet_from_file(
        infile=sample_sheet_path, flow_cell_sample_type=flow_cell_sample_type
    )
    sample_internal_ids: List[str] = []
    for sample in sample_sheet.samples:
        sample_internal_id = sample.sample_id.split("_")[0]
        if is_valid_sample_internal_id(sample_internal_id=sample_internal_id):
            sample_internal_ids.append(sample_internal_id)
    return list(set(sample_internal_ids))



"""Tests for functions related to indexes."""
from typing import Dict, List, Set, Tuple

import pytest

from cg.apps.demultiplex.sample_sheet.index import (
    INDEX_ONE_PAD_SEQUENCE,
    INDEX_TWO_PAD_SEQUENCE,
    LONG_INDEX_CYCLE_NR,
    Index,
    adapt_indexes_for_sample,
    get_hamming_distance_for_indexes,
    get_index_pair,
    get_indexes_by_lane,
    get_reagent_kit_version,
    get_reverse_complement_dna_seq,
    get_valid_indexes,
    index_exists,
    is_reverse_complement,
    update_barcode_mismatch_values_for_sample,
)
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleNovaSeq6000Bcl2Fastq,
    FlowCellSampleNovaSeqX,
)
from cg.models.demultiplex.run_parameters import RunParameters


def test_get_valid_indexes():
    """Test that the function get_valid_indexes returns a list of Index objects."""
    # GIVEN a sample sheet API

    # WHEN fetching the indexes
    indexes: List[Index] = get_valid_indexes()

    # THEN assert that the indexes are correct
    assert indexes
    assert isinstance(indexes[0], Index)


def test_index_exists_for_existing_index(valid_index: Index):
    """Test that checking the existence of an existent index returns true."""
    # GIVEN a list of indexes
    indexes: Set[str] = set(index.sequence for index in get_valid_indexes())
    # GIVEN an existent index
    existent_index: str = valid_index.sequence

    # WHEN testing for the existence of the index

    # THEN the test returns true
    assert index_exists(index=existent_index, indexes=indexes)


def test_index_exists_for_non_existent_index():
    """Test that checking the existence of a non-existent index returns false."""
    # GIVEN a list of indexes
    indexes: Set[str] = set(index.sequence for index in get_valid_indexes())
    # GIVEN a non-existent index
    non_existent_index: str = "non-existent-index"

    # WHEN testing for the existence of the index

    # THEN the test returns false
    assert not index_exists(index=non_existent_index, indexes=indexes)


def test_get_indexes_by_lane(
    novaseq6000_flow_cell_sample_1: FlowCellSampleNovaSeq6000Bcl2Fastq,
    novaseq6000_flow_cell_sample_2: FlowCellSampleNovaSeq6000Bcl2Fastq,
):
    """Test that getting indexes by lane groups indexes correctly."""
    # GIVEN two samples on different lanes
    assert novaseq6000_flow_cell_sample_1.lane != novaseq6000_flow_cell_sample_2.lane

    # WHEN getting indexes by lane
    indexes_by_lane: Dict[int, Set[str]] = get_indexes_by_lane(
        samples=[novaseq6000_flow_cell_sample_1, novaseq6000_flow_cell_sample_2]
    )

    # THEN the result dictionary has two items
    assert len(indexes_by_lane.keys()) == 2
    assert len(indexes_by_lane.values()) == 2


def test_get_reagent_kit_version_non_existent_reagent(caplog):
    """Test that getting a non-existent reagent kit version fails."""
    # GIVEN a non-existent reagent kit version
    non_existent_reagent: str = "reagent_does_not_exist"

    # WHEN getting the reagent kit version
    with pytest.raises(SyntaxError):
        get_reagent_kit_version(reagent_kit_version=non_existent_reagent)
        # THEN an error is raised
        assert caplog.text == f"Unknown reagent kit version {non_existent_reagent}"


def test_get_reagent_kit_version_all_possible_versions():
    """Test that get_reagent_kit_version works with the valid versions."""
    # GIVEN the valid reagent kit versions
    valid_versions: List[str] = ["1", "3"]
    valid_outputs: List = []

    # WHEN getting the reagent kit versions of the inputs
    for version in valid_versions:
        valid_outputs.append(get_reagent_kit_version(reagent_kit_version=version))

    # THEN the output versions are valid
    assert valid_outputs == ["1.0", "1.5"]


def test_get_index_pair_valid_cases(
    novaseq6000_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeq6000Bcl2Fastq,
):
    """Test that getting individual indexes from valid dual indexes works."""
    # GIVEN a list of dual indexes and their correct separation into individual indexes
    dual_indexes: List[str] = ["ABC-DEF", "   GHI - JKL   ", "MNO-PQR\n"]
    expected_results: List[Tuple[str, str]] = [("ABC", "DEF"), ("GHI", "JKL"), ("MNO", "PQR")]

    # WHEN getting the individual indexes of the valid indexes
    for index, expected in zip(dual_indexes, expected_results):
        novaseq6000_flow_cell_sample_before_adapt_indexes.index = index
        # THEN the result is the expected result for all cases
        assert get_index_pair(sample=novaseq6000_flow_cell_sample_before_adapt_indexes) == expected


def test_get_reverse_complement():
    """Test that getting the reverse complement of a DNA strain returns the correct sequence."""
    # GIVEN a DNA strain and its reverse complement
    dna_strain: str = "ACCTCTGT"
    reversed_complement: str = "ACAGAGGT"

    # WHEN getting the reverse complement of the DNA strain
    returned_reverse_complement: str = get_reverse_complement_dna_seq(dna=dna_strain)

    # THEN the result is the expected
    assert returned_reverse_complement == reversed_complement


def test_get_reverse_complement_not_dna(caplog):
    """Test that using a string without 'A', 'C', 'G' or 'T' fails."""
    # GIVEN a non-DNA strain
    strain: str = "ACCUCTGU"

    # WHEN getting the reverse complement
    with pytest.raises(KeyError):
        # THEN the process fails due to not recognising the unknown nucleotide
        get_reverse_complement_dna_seq(dna=strain)


def test_adapt_barcode_mismatch_values(
    lims_novaseq_x_samples: List[FlowCellSampleNovaSeqX],
    novaseq_x_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeqX,
):
    """Test that the barcode mismatch values are updated for a sample."""
    # GIVEN a list of NovaSeqX samples
    present_index: str = lims_novaseq_x_samples[0].index
    # GIVEN a sample
    novaseq_x_flow_cell_sample_before_adapt_indexes.index = present_index
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_1 == 1
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_2 == 1

    # WHEN adapting the barcode mismatch values
    update_barcode_mismatch_values_for_sample(
        sample_to_update=novaseq_x_flow_cell_sample_before_adapt_indexes,
        samples_to_compare_to=lims_novaseq_x_samples,
    )

    # THEN the barcode mismatch values have been updated
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_1 == 0
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_2 == 0


def test_adapt_barcode_mismatch_values_repeated_sample(
    novaseq_x_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeqX,
):
    """Test that a sample does not compare to itself when adapting barcode mismatching values."""
    # GIVEN a list of repeated unadapted NovaSeqX samples
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_1 == 1
    assert novaseq_x_flow_cell_sample_before_adapt_indexes.barcode_mismatches_2 == 1
    samples: List[FlowCellSampleNovaSeqX] = [
        novaseq_x_flow_cell_sample_before_adapt_indexes for _ in range(3)
    ]

    # WHEN adapting the barcode mismatch values for the samples
    update_barcode_mismatch_values_for_sample(
        sample_to_update=novaseq_x_flow_cell_sample_before_adapt_indexes,
        samples_to_compare_to=samples,
    )

    # THEN the barcode mismatch values remains being 1 for all samples
    for sample in samples:
        assert sample.barcode_mismatches_1 == 1
        assert sample.barcode_mismatches_2 == 1


def test_adapt_indexes_for_sample_reverse_complement_padding(
    novaseq_6000_run_parameters: RunParameters,
    novaseq6000_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeq6000Bcl2Fastq,
):
    """Test that adapting indexes for a sample that needs padding and reverse complement works."""
    # GIVEN a run parameters file that needs reverse complement of indexes
    assert is_reverse_complement(run_parameters=novaseq_6000_run_parameters)
    # GIVEN a sample that needs padding
    sample: FlowCellSampleNovaSeq6000Bcl2Fastq = novaseq6000_flow_cell_sample_before_adapt_indexes
    assert novaseq_6000_run_parameters.get_index_1_cycles() == LONG_INDEX_CYCLE_NR
    sample.index = "ATTCCACA-TGGTCTTG"

    # WHEN adapting the indexes of the sample
    adapt_indexes_for_sample(
        sample=sample,
        index_cycles=novaseq_6000_run_parameters.index_length,
        reverse_complement=is_reverse_complement(run_parameters=novaseq_6000_run_parameters),
    )

    # THEN the first index was correctly adapted
    assert len(sample.index) == LONG_INDEX_CYCLE_NR
    assert sample.index[-2:] == INDEX_ONE_PAD_SEQUENCE
    # THEN the second index was correctly adapted
    assert len(sample.index2) == LONG_INDEX_CYCLE_NR
    assert sample.index2[-2:] == get_reverse_complement_dna_seq(dna=INDEX_TWO_PAD_SEQUENCE)


def test_adapt_indexes_for_sample_reverse_complement_no_padding(
    novaseq_6000_run_parameters: RunParameters,
    novaseq6000_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeq6000Bcl2Fastq,
):
    """Test that adapting indexes of a sample that needs reverse complement but no padding works."""
    # GIVEN a run parameters file that needs reverse complement of indexes
    assert is_reverse_complement(run_parameters=novaseq_6000_run_parameters)
    # GIVEN a sample that does not need padding
    sample: FlowCellSampleNovaSeq6000Bcl2Fastq = novaseq6000_flow_cell_sample_before_adapt_indexes
    assert novaseq_6000_run_parameters.get_index_1_cycles() == LONG_INDEX_CYCLE_NR
    assert len(sample.index) >= 2 * LONG_INDEX_CYCLE_NR
    initial_indexes: List[str] = sample.index.split("-")
    initial_index1: str = initial_indexes[0]
    initial_index2: str = initial_indexes[1]

    # WHEN adapting the indexes of the sample
    adapt_indexes_for_sample(
        sample=sample,
        index_cycles=novaseq_6000_run_parameters.index_length,
        reverse_complement=is_reverse_complement(run_parameters=novaseq_6000_run_parameters),
    )

    # THEN the first index was correctly adapted
    assert len(sample.index) == LONG_INDEX_CYCLE_NR
    assert sample.index == initial_index1
    # THEN the second index was correctly adapted
    assert len(sample.index2) == LONG_INDEX_CYCLE_NR
    assert sample.index2 == get_reverse_complement_dna_seq(dna=initial_index2)


def test_adapt_indexes_for_sample_no_reverse_complement_no_padding(
    novaseq_x_run_parameters: RunParameters,
    novaseq_x_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeqX,
):
    """Test that adapting indexes of a sample that does not need reverse complement nor padding works."""
    # GIVEN a run parameters file that does not need reverse complement of indexes
    assert not is_reverse_complement(run_parameters=novaseq_x_run_parameters)
    # GIVEN a sample that does not need padding
    sample: FlowCellSampleNovaSeqX = novaseq_x_flow_cell_sample_before_adapt_indexes
    assert novaseq_x_run_parameters.get_index_1_cycles() == LONG_INDEX_CYCLE_NR
    assert len(sample.index) >= 2 * LONG_INDEX_CYCLE_NR
    initial_indexes: List[str] = sample.index.split("-")
    initial_index1: str = initial_indexes[0]
    initial_index2: str = initial_indexes[1]

    # WHEN adapting the indexes of the sample
    adapt_indexes_for_sample(
        sample=sample,
        index_cycles=novaseq_x_run_parameters.index_length,
        reverse_complement=is_reverse_complement(run_parameters=novaseq_x_run_parameters),
    )

    # THEN the first index was correctly adapted
    assert len(sample.index) == LONG_INDEX_CYCLE_NR
    assert sample.index == initial_index1
    # THEN the second index was correctly adapted
    assert len(sample.index2) == LONG_INDEX_CYCLE_NR
    assert sample.index2 == initial_index2


def test_get_hamming_distance_index_different_lengths():
    """Test that the hamming distance between indexes with same prefix but different lengths is zero."""
    # GIVEN two index_1 sequences with the same prefixes but different lengths
    sequence_1: str = "GATTACA"
    sequence_2: str = "GATTACAXX"

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is zero
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_2) == 0
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_1) == 0

    # WHEN getting the hamming distance between themselves

    # THEN the distance is zero
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_1) == 0
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_2) == 0


def test_get_hamming_distance_index_different_prefixes():
    """Test that the hamming distance between indexes with different prefixes is greater than zero."""
    # GIVEN two index_1 sequences with the same suffixes but different prefixes
    sequence_1: str = "GATTACA"
    sequence_2: str = "XXGATTACA"

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is greater than zero
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_2) > 0
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_1) > 0



"""Tests for lims functions that are related with sample sheets."""

from cg.apps.lims.sample_sheet import extract_sequence_in_parentheses


def test_extract_sequence_in_parentheses(reagent_label: str, reagent_sequence: str):
    """Test that the correct index is fetched from the reagent label."""
    # GIVEN a reagent with compatible label and sequence

    # WHEN getting the index
    extracted_sequence: str = extract_sequence_in_parentheses(label=reagent_label)

    # THEN the index is what is expected
    assert extracted_sequence == reagent_sequence


def test_extract_sequence_in_parentheses_invalid_reagent(
    invalid_reagent_label: str, invalid_reagent_sequence: str
):
    """Test that the extracted index is not the same as the reagent sequence."""
    # GIVEN a reagent with compatible label and sequence

    # WHEN getting the index
    extracted_sequence: str = extract_sequence_in_parentheses(label=invalid_reagent_label)

    # THEN the index is what is expected
    assert extracted_sequence != invalid_reagent_sequence


def test_extract_sequence_in_parentheses_no_parentheses(label_no_parentheses: str):
    """Test using a label without parentheses returns None."""
    # GIVEN a reagent with compatible label and sequence

    # WHEN getting the index
    extracted_sequence: str = extract_sequence_in_parentheses(label=label_no_parentheses)

    # THEN the index is what is expected
    assert extracted_sequence is None



import shutil
from pathlib import Path

from cg.cli.workflow.commands import link
from cg.constants import EXIT_SUCCESS
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.models.cg_config import CGConfig
from click.testing import CliRunner


def test_cli_link_no_case(
    cli_runner: CliRunner,
    fluffy_case_id_non_existing,
    fluffy_sample_lims_id,
    fluffy_context: CGConfig,
    caplog,
):
    caplog.set_level("ERROR")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does not exist in database

    # WHEN running command in dry-run
    result = cli_runner.invoke(link, [fluffy_case_id_non_existing], obj=fluffy_context)

    # THEN command does not terminate successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN log informs that case id does not exist
    assert fluffy_case_id_non_existing in caplog.text

    # THEN file directories were not created
    assert not Path(
        fluffy_analysis_api.get_fastq_path(
            case_id=fluffy_case_id_non_existing, sample_id=fluffy_sample_lims_id
        )
    ).exists()


def test_cli_link(
    cli_runner: CliRunner,
    fluffy_case_id_existing,
    fluffy_sample_lims_id,
    fluffy_context: CGConfig,
    fastq_file_fixture_path,
    caplog,
):
    caplog.set_level("INFO")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]
    # GIVEN that a fastq path does not exist
    fastq_path: Path = Path(
        fluffy_analysis_api.get_fastq_path(
            case_id=fluffy_case_id_existing, sample_id=fluffy_sample_lims_id
        )
    )
    assert not fastq_path.exists()

    # GIVEN a case_id that does exist in database

    # WHEN running command
    result = cli_runner.invoke(link, [fluffy_case_id_existing], obj=fluffy_context)

    # THEN log informs about what paths it links from and to
    assert "Linking" in caplog.text

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN file directories were created
    assert fastq_path.exists()

    # Clean-uo
    shutil.rmtree(fastq_path)


def test_cli_link_dir_exists(
    cli_runner: CliRunner,
    fluffy_case_id_existing,
    fluffy_sample_lims_id,
    fluffy_context: CGConfig,
    fastq_file_fixture_path,
    caplog,
):
    caplog.set_level("INFO")
    fluffy_analysis_api: FluffyAnalysisAPI = fluffy_context.meta_apis["analysis_api"]

    # GIVEN a case_id that does exist in database

    # GIVEN folder with fastq files already present in working directory
    Path(fluffy_analysis_api.get_workdir_path(fluffy_case_id_existing)).mkdir(
        exist_ok=True, parents=True
    )

    # WHEN running command
    result = cli_runner.invoke(link, [fluffy_case_id_existing], obj=fluffy_context)

    # THEN log informs about what paths it links from and to
    assert "Linking" in caplog.text

    # THEN command terminates successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN file directories were created
    assert Path(
        fluffy_analysis_api.get_fastq_path(
            case_id=fluffy_case_id_existing, sample_id=fluffy_sample_lims_id
        )
    ).exists()

    # THEN log informs that directory exists and will be re-created
    assert "directory exists" in caplog.text



from typing import Optional, List, Callable
from enum import Enum
from sqlalchemy.orm import Query

from cg.store.models import Family, Sample


def get_samples_in_case_by_internal_id(
    case_samples: Query, case_internal_id: str, **kwargs
) -> Query:
    """Return samples associated with a case."""
    return case_samples.filter(Family.internal_id == case_internal_id)


def get_cases_with_sample_by_internal_id(case_samples: Query, sample_internal_id: str, **kwargs):
    """Return cases associated with a sample internal id."""
    return case_samples.filter(Sample.internal_id == sample_internal_id)


def apply_case_sample_filter(
    filter_functions: List[Callable],
    case_samples: Query,
    case_internal_id: Optional[str] = None,
    sample_entry_id: Optional[int] = None,
    sample_internal_id: Optional[str] = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""

    for function in filter_functions:
        case_samples: Query = function(
            case_samples=case_samples,
            case_internal_id=case_internal_id,
            sample_entry_id=sample_entry_id,
            sample_internal_id=sample_internal_id,
        )
    return case_samples


class CaseSampleFilter(Enum):
    """Define CaseSample filter functions."""

    GET_SAMPLES_IN_CASE_BY_INTERNAL_ID: Callable = get_samples_in_case_by_internal_id
    GET_CASES_WITH_SAMPLE_BY_INTERNAL_ID: Callable = get_cases_with_sample_by_internal_id



import datetime as dt
import pytest

from typing import List
from cg.constants.constants import PrepCategory
from cg.constants.priority import PriorityTerms
from cg.meta.orders.pool_submitter import PoolSubmitter
from cg.store import Store
from tests.meta.demultiplex.conftest import fixture_populated_flow_cell_store
from tests.store_helpers import StoreHelpers
from cg.constants.invoice import CustomerNames
from cg.constants.subject import PhenotypeStatus
from cg.constants import Pipeline


@pytest.fixture(name="microbial_store")
def fixture_microbial_store(store: Store, helpers: StoreHelpers) -> Store:
    """Populate a store with microbial application tags"""
    microbial_active_apptags = ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"]
    microbial_inactive_apptags = ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"]

    for app_tag in microbial_active_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=False)

    for app_tag in microbial_inactive_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=True)

    return store


@pytest.fixture(name="rml_pool_store")
def fixture_rml_pool_store(
    case_id: str,
    customer_id: str,
    helpers,
    sample_id: str,
    store: Store,
    ticket_id: str,
    timestamp_now: dt.datetime,
):
    new_customer = store.add_customer(
        internal_id=customer_id,
        name="Test customer",
        scout_access=True,
        invoice_address="skolgatan 15",
        invoice_reference="abc",
    )
    store.session.add(new_customer)

    application = store.add_application(
        tag="RMLP05R800",
        prep_category="rml",
        description="Ready-made",
        percent_kth=80,
        percent_reads_guaranteed=75,
        sequencing_depth=0,
        target_reads=800,
    )
    store.session.add(application)

    app_version = store.add_application_version(
        application=application,
        version=1,
        valid_from=timestamp_now,
        prices={
            PriorityTerms.STANDARD: 12,
            PriorityTerms.PRIORITY: 222,
            PriorityTerms.EXPRESS: 123,
            PriorityTerms.RESEARCH: 12,
        },
    )
    store.session.add(app_version)

    new_pool = store.add_pool(
        customer=new_customer,
        name="Test",
        order="Test",
        ordered=dt.datetime.now(),
        application_version=app_version,
    )
    store.session.add(new_pool)
    new_case = helpers.add_case(
        store=store,
        internal_id=case_id,
        name=PoolSubmitter.create_case_name(ticket=ticket_id, pool_name="Test"),
    )
    store.session.add(new_case)

    new_sample = helpers.add_sample(
        store=store,
        internal_id=sample_id,
        application_tag=application.tag,
        application_type=application.prep_category,
        customer_id=new_customer.id,
    )
    new_sample.application_version = app_version
    store.session.add(new_sample)
    store.session.commit()

    helpers.add_relationship(
        store=store,
        sample=new_sample,
        case=new_case,
    )

    yield store


@pytest.fixture(name="re_sequenced_sample_store")
def fixture_re_sequenced_sample_store(
    store: Store,
    dragen_flow_cell_id: str,
    case_id: str,
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    ticket_id: str,
    timestamp_now: dt.datetime,
    helpers,
) -> Store:
    """Populate a store with a Fluffy case, with a sample that has been sequenced on two flow cells."""
    re_sequenced_sample_store: Store = store
    store_case = helpers.add_case(
        store=re_sequenced_sample_store,
        internal_id=case_id,
        name=family_name,
        data_analysis=Pipeline.FLUFFY,
    )

    store_sample = helpers.add_sample(
        internal_id=sample_id,
        is_tumour=False,
        application_type=PrepCategory.READY_MADE_LIBRARY.value,
        reads=1200000000,
        store=re_sequenced_sample_store,
        original_ticket=ticket_id,
        sequenced_at=timestamp_now,
    )

    one_day_ahead_of_now = timestamp_now + dt.timedelta(days=1)

    helpers.add_flowcell(
        store=re_sequenced_sample_store,
        flow_cell_name=dragen_flow_cell_id,
        samples=[store_sample],
        date=timestamp_now,
    )

    helpers.add_flowcell(
        store=re_sequenced_sample_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        samples=[store_sample],
        date=one_day_ahead_of_now,
    )

    helpers.add_relationship(store=re_sequenced_sample_store, case=store_case, sample=store_sample)

    return re_sequenced_sample_store


@pytest.fixture(name="max_nr_of_cases")
def fixture_max_nr_of_cases() -> int:
    """Return the number of maximum number of cases"""
    return 50


@pytest.fixture(name="max_nr_of_samples")
def fixture_max_nr_of_samples() -> int:
    """Return the number of maximum number of samples"""
    return 50


@pytest.fixture(name="EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS")
def fixture_expected_number_of_not_archived_applications() -> int:
    """Return the number of expected number of not archived applications"""
    return 4


@pytest.fixture(name="EXPECTED_NUMBER_OF_APPLICATIONS")
def fixture_expected_number_of_applications() -> int:
    """Return the number of expected number of applications with prep category"""
    return 7


@pytest.fixture(name="store_with_samples_that_have_names")
def store_with_samples_that_have_names(
    store: Store, helpers: StoreHelpers, name="sample_1"
) -> Store:
    """Return a store with two samples of which one has a name"""
    for index in range(1, 4):
        helpers.add_sample(
            store=store, internal_id=f"test_sample_{index}", name=f"test_sample_{index}"
        )

    helpers.add_sample(
        store=store,
        internal_id="unrelated_id",
        name="unrelated_name",
        customer_id="unrelated_customer",
    )
    return store


@pytest.fixture(name="cust123")
def fixture_cust123() -> str:
    """Return a customer id"""
    return "cust123"


@pytest.fixture(name="test_subject")
def fixture_test_subject() -> str:
    """Return a subject id"""
    return "test_subject"


@pytest.fixture(name="store_with_samples_subject_id_and_tumour_status")
def fixture_store_with_samples_subject_id_and_tumour_status(
    store: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
) -> Store:
    """Return a store with two samples that have subject ids of which one is tumour"""
    helpers.add_sample(
        store=store,
        internal_id="test_sample_1",
        name="sample_1",
        subject_id=test_subject,
        is_tumour=True,
        customer_id=cust123,
    )

    helpers.add_sample(
        store=store,
        internal_id="test_sample_2",
        name="sample_2",
        subject_id=test_subject,
        is_tumour=False,
        customer_id=cust123,
    )
    return store


@pytest.fixture(name="store_with_samples_and_tumour_status_missing_subject_id")
def fixture_store_with_samples_and_tumour_status_missing_subject_id(
    store: Store,
    helpers: StoreHelpers,
    cust123: str,
    test_subject: str,
) -> Store:
    """Return a store with two samples of which one is tumour"""
    helpers.add_sample(
        store=store,
        internal_id="test_sample_1",
        name="sample_1",
        is_tumour=True,
        customer_id=cust123,
    )

    helpers.add_sample(
        store=store,
        internal_id="test_sample_2",
        name="sample_2",
        is_tumour=False,
        customer_id=cust123,
    )
    return store


@pytest.fixture(name="store_with_samples_customer_id_and_subject_id_and_tumour_status")
def store_with_samples_customer_id_and_subject_id_and_tumour_status(
    store: Store, helpers: StoreHelpers
) -> Store:
    """Return a store with four samples with different customer IDs, and tumour status."""
    samples_data = [
        # customer_id, subject_id, is_tumour
        ("1", "test_subject", True),
        ("1", "test_subject_2", False),
        ("2", "test_subject", True),
        ("2", "test_subject_2", False),
    ]
    for customer_id, subject_id, is_tumour in samples_data:
        helpers.add_sample(
            store=store,
            internal_id=f"test_sample_{customer_id}_{subject_id}",
            name=f"sample_{customer_id}_{subject_id}",
            subject_id=subject_id,
            is_tumour=is_tumour,
            customer_id=customer_id,
        )
    return store


@pytest.fixture(name="pool_name_1")
def fixture_pool_name_1() -> str:
    """Return the name of the first pool."""
    return "pool_1"


@pytest.fixture(name="pool_order_1")
def fixture_pool_order_1() -> str:
    """Return the order of the first pool."""
    return "pool_order_1"


@pytest.fixture(name="store_with_multiple_pools_for_customer")
def fixture_store_with_multiple_pools_for_customer(
    store: Store,
    helpers: StoreHelpers,
    customer_id: str = CustomerNames.cust132,
) -> Store:
    """Return a store with two pools with different names for the same customer."""
    for number in range(2):
        helpers.ensure_pool(
            store=store,
            customer_id=customer_id,
            name="_".join(["pool", str(number)]),
            order="_".join(["pool", "order", str(number)]),
        )
    yield store


@pytest.fixture(name="store_with_active_sample_analyze")
def fixture_store_with_active_sample_analyze(store: Store, helpers: StoreHelpers) -> Store:
    """Return a store with an active sample with action analyze."""
    # GIVEN a store with a sample that is active
    case = helpers.add_case(
        store=store, name="test_case", internal_id="test_case_internal_id", action="analyze"
    )
    sample = helpers.add_sample(
        store=store, name="test_sample", internal_id="test_sample_internal_id"
    )
    helpers.add_relationship(store=store, sample=sample, case=case)

    yield store


@pytest.fixture(name="store_with_active_sample_running")
def fixture_store_with_active_sample_running(store: Store, helpers: StoreHelpers) -> Store:
    """Return a store with an active sample with action running."""
    # GIVEN a store with a sample that is active
    case = helpers.add_case(
        store=store, name="test_case", internal_id="test_case_internal_id", action="running"
    )
    sample = helpers.add_sample(
        store=store, name="test_sample", internal_id="test_sample_internal_id"
    )
    helpers.add_relationship(store=store, sample=sample, case=case)

    yield store


@pytest.fixture(name="three_customer_ids")
def fixture_three_customer_ids() -> List[str]:
    """Return three customer ids."""
    yield ["".join(["cust00", str(number)]) for number in range(3)]


@pytest.fixture(name="three_pool_names")
def fixture_three_pool_names() -> List[str]:
    """Return three customer ids."""
    yield ["_".join(["test_pool", str(number)]) for number in range(3)]


@pytest.fixture(name="store_with_samples_for_multiple_customers")
def fixture_store_with_samples_for_multiple_customers(
    store: Store, helpers: StoreHelpers, timestamp_now: dt.datetime
) -> Store:
    """Return a store with two samples for three different customers."""
    for number in range(3):
        helpers.add_sample(
            store=store,
            internal_id="_".join(["test_sample", str(number)]),
            customer_id="".join(["cust00", str(number)]),
            no_invoice=False,
            delivered_at=timestamp_now,
        )
    yield store


@pytest.fixture(name="store_with_pools_for_multiple_customers")
def fixture_store_with_pools_for_multiple_customers(
    store: Store, helpers: StoreHelpers, timestamp_now: dt.datetime
) -> Store:
    """Return a store with two samples for three different customers."""
    for number in range(3):
        helpers.ensure_pool(
            store=store,
            name="_".join(["test_pool", str(number)]),
            customer_id="".join(["cust00", str(number)]),
            no_invoice=False,
            delivered_at=timestamp_now,
        )
    yield store


@pytest.fixture(name="store_with_analyses_for_cases")
def fixture_store_with_analyses_for_cases(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_yesterday,
            completed_at=timestamp_yesterday,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            uploaded_at=timestamp_now,
            delivery_reported_at=None,
            uploaded_to_vogue_at=None,
            completed_at=timestamp_now,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        analysis_store.relate_sample(
            family=oldest_analysis.family, sample=sample, status=PhenotypeStatus.UNKNOWN
        )

    return analysis_store


@pytest.fixture(name="store_with_analyses_for_cases_not_uploaded_fluffy")
def fixture_store_with_analyses_for_cases_not_uploaded_fluffy(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases and pipeline."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_yesterday,
            pipeline=Pipeline.FLUFFY,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            uploaded_at=None,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_now,
            pipeline=Pipeline.FLUFFY,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        analysis_store.relate_sample(
            family=oldest_analysis.family, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
    return analysis_store


@pytest.fixture(name="store_with_analyses_for_cases_not_uploaded_microsalt")
def fixture_store_with_analyses_for_cases_not_uploaded_microsalt(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases and pipeline."""

    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            uploaded_at=timestamp_yesterday,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_yesterday,
            pipeline=Pipeline.MICROSALT,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            uploaded_at=None,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_now,
            pipeline=Pipeline.MICROSALT,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
        analysis_store.relate_sample(
            family=oldest_analysis.family, sample=sample, status=PhenotypeStatus.UNKNOWN
        )
    return analysis_store


@pytest.fixture(name="store_with_analyses_for_cases_to_deliver")
def fixture_store_with_analyses_for_cases_to_deliver(
    analysis_store: Store,
    helpers: StoreHelpers,
    timestamp_now: dt.datetime,
    timestamp_yesterday: dt.datetime,
) -> Store:
    """Return a store with two analyses for two cases."""
    case_one = analysis_store.get_case_by_internal_id("yellowhog")
    case_two = helpers.add_case(analysis_store, internal_id="test_case_1")

    cases = [case_one, case_two]
    for case in cases:
        oldest_analysis = helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_yesterday,
            uploaded_at=None,
            delivery_reported_at=None,
            uploaded_to_vogue_at=timestamp_yesterday,
            completed_at=timestamp_yesterday,
            pipeline=Pipeline.FLUFFY,
        )
        helpers.add_analysis(
            analysis_store,
            case=case,
            started_at=timestamp_now,
            uploaded_at=None,
            delivery_reported_at=None,
            uploaded_to_vogue_at=None,
            completed_at=timestamp_now,
            pipeline=Pipeline.MIP_DNA,
        )
        sample = helpers.add_sample(analysis_store, delivered_at=None)
        analysis_store.relate_sample(
            family=oldest_analysis.family, sample=sample, status=PhenotypeStatus.UNKNOWN
        )

    return analysis_store



import logging
from copy import deepcopy

from cg.constants import DEFAULT_CAPTURE_KIT
from cg.constants.subject import RelationshipStatus
from cg.exc import PedigreeConfigError
from marshmallow import Schema, fields, validate

LOG = logging.getLogger(__name__)


class SampleSchema(Schema):
    sample_id = fields.Str(required=True)
    sample_display_name = fields.Str()
    analysis_type = fields.Str(
        required=True,
        validate=validate.OneOf(
            choices=[
                "tga",
                "wes",
                "wgs",
                "wts",
            ]
        ),
    )
    father = fields.Str(dump_default=RelationshipStatus.HAS_NO_PARENT.value)
    mother = fields.Str(dump_default=RelationshipStatus.HAS_NO_PARENT.value)
    phenotype = fields.Str(
        required=True,
        validate=validate.OneOf(choices=["affected", "unaffected", "unknown"]),
    )
    sex = fields.Str(required=True, validate=validate.OneOf(choices=["female", "male", "unknown"]))
    expected_coverage = fields.Float()
    capture_kit = fields.Str(dump_default=DEFAULT_CAPTURE_KIT)


class ConfigSchema(Schema):
    case = fields.Str(required=True)
    default_gene_panels = fields.List(fields.Str(), required=True)
    samples = fields.List(fields.Nested(SampleSchema), required=True)


class ConfigHandler:
    @staticmethod
    def make_pedigree_config(data: dict) -> dict:
        """Make a MIP pedigree config"""
        ConfigHandler.validate_config(data=data)
        return ConfigHandler.parse_pedigree_config(data)

    @staticmethod
    def validate_config(data: dict) -> dict:
        """Validate MIP pedigree config format"""
        errors = ConfigSchema().validate(data)
        fatal_error = False
        for field, messages in errors.items():
            if isinstance(messages, dict):
                for sample_index, sample_errors in messages.items():
                    try:
                        sample_id = data["samples"][sample_index]["sample_id"]
                    except KeyError:
                        raise PedigreeConfigError("missing sample id")
                    for sample_key, sub_messages in sample_errors.items():
                        if sub_messages != ["Unknown field."]:
                            fatal_error = True
                        LOG.error(f"{sample_id} -> {sample_key}: {', '.join(sub_messages)}")
            else:
                fatal_error = True
                LOG.error(f"{field}: {', '.join(messages)}")
        if fatal_error:
            raise PedigreeConfigError("invalid config input", errors=errors)
        return errors

    @staticmethod
    def parse_pedigree_config(data: dict) -> dict:
        """Parse the pedigree config data"""
        data_copy = deepcopy(data)
        # handle single sample cases with 'unknown' phenotype
        if len(data_copy["samples"]) == 1 and data_copy["samples"][0]["phenotype"] == "unknown":
            LOG.info("setting 'unknown' phenotype to 'unaffected'")
            data_copy["samples"][0]["phenotype"] = "unaffected"
        for sample_data in data_copy["samples"]:
            sample_data["mother"] = (
                sample_data.get("mother") or RelationshipStatus.HAS_NO_PARENT.value
            )
            sample_data["father"] = (
                sample_data.get("father") or RelationshipStatus.HAS_NO_PARENT.value
            )
            if sample_data["analysis_type"] == "wgs" and sample_data.get("capture_kit") is None:
                sample_data["capture_kit"] = DEFAULT_CAPTURE_KIT
        return data_copy



from datetime import date
from typing import Optional

from pydantic.v1 import BaseModel, conlist
from typing_extensions import Literal


class CrunchyFile(BaseModel):
    path: str
    file: Literal["first_read", "second_read", "spring"]
    checksum: Optional[str]
    algorithm: Literal["sha1", "md5", "sha256"] = None
    updated: date = None


class CrunchyMetadata(BaseModel):
    files: conlist(CrunchyFile, max_items=3, min_items=3)



"""Module for Taxprofiler Analysis API."""

import logging
from typing import Dict, List, Optional

from pydantic.v1 import ValidationError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.cg_config import CGConfig
from cg.constants import Pipeline
from cg.constants.nextflow import NFX_READ1_HEADER, NFX_READ2_HEADER, NFX_SAMPLE_HEADER
from cg.constants.sequencing import SequencingPlatform
from cg.constants.taxprofiler import (
    TAXPROFILER_INSTRUMENT_PLATFORM,
    TAXPROFILER_RUN_ACCESSION,
    TAXPROFILER_SAMPLE_SHEET_HEADERS,
    TAXPROFILER_FASTA_HEADER,
)
from cg.meta.workflow.fastq import TaxprofilerFastqHandler
from cg.meta.workflow.nextflow_common import NextflowAnalysisAPI
from cg.models.taxprofiler.taxprofiler_sample import TaxprofilerSample
from cg.store.models import Family, Sample

LOG = logging.getLogger(__name__)


class TaxprofilerAnalysisAPI(AnalysisAPI):
    """Handles communication between Taxprofiler processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.TAXPROFILER,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir: str = config.taxprofiler.root

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def fastq_handler(self):
        return TaxprofilerFastqHandler

    def get_case_config_path(self, case_id):
        return NextflowAnalysisAPI.get_case_config_path(case_id=case_id, root_dir=self.root_dir)

    @staticmethod
    def build_sample_sheet_content(
        sample_name: str,
        fastq_r1: List[str],
        fastq_r2: List[str],
        instrument_platform: SequencingPlatform.ILLUMINA,
        fasta: Optional[str] = "",
    ) -> Dict[str, List[str]]:
        """Build sample sheet headers and lists."""
        try:
            TaxprofilerSample(
                sample=sample_name,
                fastq_r1=fastq_r1,
                fastq_r2=fastq_r2,
                instrument_platform=instrument_platform,
            )
        except ValidationError as error:
            LOG.error(error)
            raise ValueError

        # Complete sample lists to the same length as fastq_r1:
        samples_full_list: List[str] = [sample_name] * len(fastq_r1)
        instrument_full_list: List[str] = [instrument_platform] * len(fastq_r1)
        fasta_full_list: List[str] = [fasta] * len(fastq_r1)

        sample_sheet_content: Dict[str, List[str]] = {
            NFX_SAMPLE_HEADER: samples_full_list,
            TAXPROFILER_RUN_ACCESSION: samples_full_list,
            TAXPROFILER_INSTRUMENT_PLATFORM: instrument_full_list,
            NFX_READ1_HEADER: fastq_r1,
            NFX_READ2_HEADER: fastq_r2,
            TAXPROFILER_FASTA_HEADER: fasta_full_list,
        }

        return sample_sheet_content

    def write_sample_sheet(
        self,
        case_id: str,
        instrument_platform: SequencingPlatform.ILLUMINA,
        fasta: Optional[str],
    ) -> None:
        """Write sample sheet for taxprofiler analysis in case folder."""
        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content: Dict[str, List[str]] = {}

        for link in case.links:
            sample_name: str = link.sample.name
            sample_metadata: List[str] = self.gather_file_metadata_for_sample(link.sample)
            fastq_r1: List[str] = NextflowAnalysisAPI.extract_read_files(
                read_nb=1, metadata=sample_metadata
            )
            fastq_r2: List[str] = NextflowAnalysisAPI.extract_read_files(
                read_nb=2, metadata=sample_metadata
            )
            sample_content: Dict[str, List[str]] = self.build_sample_sheet_content(
                sample_name=sample_name,
                fastq_r1=fastq_r1,
                fastq_r2=fastq_r2,
                instrument_platform=instrument_platform,
                fasta=fasta,
            )

            for headers, contents in sample_content.items():
                sample_sheet_content.setdefault(headers, []).extend(contents)

            LOG.info(sample_sheet_content)
            NextflowAnalysisAPI.create_samplesheet_csv(
                samplesheet_content=sample_sheet_content,
                headers=TAXPROFILER_SAMPLE_SHEET_HEADERS,
                config_path=NextflowAnalysisAPI.get_case_config_path(
                    case_id=case_id, root_dir=self.root_dir
                ),
            )

    def config_case(
        self,
        case_id: str,
        instrument_platform: SequencingPlatform.ILLUMINA,
        fasta: Optional[str],
    ) -> None:
        """Create sample sheet file for Taxprofiler analysis."""
        NextflowAnalysisAPI.make_case_folder(case_id=case_id, root_dir=self.root_dir)
        LOG.info("Generating sample sheet")
        self.write_sample_sheet(
            case_id=case_id,
            instrument_platform=instrument_platform,
            fasta=fasta,
        )
        LOG.info("Sample sheet written")



from datetime import datetime, timedelta
from pathlib import Path
from unittest import mock

import pytest
from requests import Response

from cg.constants.constants import FileFormat
from cg.io.controller import WriteStream
from cg.meta.archive.ddn_dataflow import DDNDataFlowApi, TransferData, TransferPayload, ROOT_TO_TRIM
from cg.models.cg_config import DDNDataFlowConfig


@pytest.fixture(name="ddn_dataflow_config")
def fixture_ddn_dataflow_config(
    local_storage_repository: str, remote_storage_repository: str
) -> DDNDataFlowConfig:
    """Returns a mock DDN Dataflow config."""
    return DDNDataFlowConfig(
        database_name="test_db",
        user="test_user",
        password="DummyPassword",
        url=Path("some", "api", "url.com").as_posix(),
        local_storage=local_storage_repository,
        archive_repository=remote_storage_repository,
    )


@pytest.fixture(name="ddn_dataflow_api")
def fixture_ddn_dataflow_api(ddn_dataflow_config: DDNDataFlowConfig) -> DDNDataFlowApi:
    """Returns a DDNApi without tokens being set."""
    mock_ddn_auth_success_response = Response()
    mock_ddn_auth_success_response.status_code = 200
    mock_ddn_auth_success_response._content = WriteStream.write_stream_from_content(
        file_format=FileFormat.JSON,
        content={
            "access": "test_auth_token",
            "refresh": "test_refresh_token",
            "expire": (datetime.now() + timedelta(minutes=20)).timestamp(),
        },
    ).encode()
    with mock.patch(
        "cg.meta.archive.ddn_dataflow.APIRequest.api_request_from_content",
        return_value=mock_ddn_auth_success_response,
    ):
        return DDNDataFlowApi(ddn_dataflow_config)


@pytest.fixture(name="transfer_data")
def fixture_transfer_data(local_directory: Path, remote_path: Path) -> TransferData:
    """Return a TransferData object."""
    return TransferData(source=local_directory.as_posix(), destination=remote_path.as_posix())


@pytest.fixture(name="transfer_payload")
def fixture_transfer_payload(transfer_data: TransferData) -> TransferPayload:
    """Return a TransferPayload object containing two identical TransferData object.."""
    return TransferPayload(files_to_transfer=[transfer_data, transfer_data.copy(deep=True)])


@pytest.fixture(name="remote_path")
def fixture_remote_path() -> Path:
    """Returns a mock path."""
    return Path("/some", "place")


@pytest.fixture(name="local_directory")
def fixture_local_directory() -> Path:
    """Returns a mock path with /home as its root."""
    return Path(ROOT_TO_TRIM, "other", "place")


@pytest.fixture(name="trimmed_local_directory")
def fixture_trimmed_local_directory(local_directory: Path) -> Path:
    """Returns the trimmed local directory."""
    return Path(f"/{local_directory.relative_to(ROOT_TO_TRIM)}")


@pytest.fixture(name="local_storage_repository")
def fixture_local_storage_repository() -> str:
    """Returns a local storage repository."""
    return "local@storage:"


@pytest.fixture(name="remote_storage_repository")
def fixture_remote_storage_repository() -> str:
    """Returns a remote storage repository."""
    return "archive@repisitory:"


@pytest.fixture(name="full_remote_path")
def fixture_full_remote_path(remote_storage_repository: str, remote_path: Path) -> str:
    """Returns the merged remote repository and path."""
    return remote_storage_repository + remote_path.as_posix()


@pytest.fixture(name="full_local_path")
def fixture_full_local_path(local_storage_repository: str, trimmed_local_directory: Path) -> str:
    """Returns the merged local repository and trimmed path."""
    return local_storage_repository + trimmed_local_directory.as_posix()



from cg.apps.sequencing_metrics_parser.sequencing_metrics_calculator import (
    q30_ratio,
    average_quality_score,
)


def test_calculate_bases_with_q30_ratio():
    """Test calculating q30 ratio."""
    # GIVEN a number of bases with q30 and a total number of bases
    q30_yield: int = 100
    total_yield: int = 200

    # WHEN calculating the q30 ratio
    q30_ratio_metric = q30_ratio(q30_yield=q30_yield, total_yield=total_yield)

    # THEN the q30 ratio should be the number of bases with q30 divided by the total number of bases
    assert q30_ratio_metric == q30_yield / total_yield


def test_calculate_lane_mean_quality_score():
    """Test calculating the average quality score."""
    # GIVEN a total quality score and a total yield
    total_quality_score: int = 100
    total_yield: int = 200

    # WHEN calculating the average quality score
    avg_quality_score: float = average_quality_score(
        total_quality_score=total_quality_score, total_yield=total_yield
    )

    # THEN the average quality score should be the total quality score divided by the total yield
    assert avg_quality_score == total_quality_score / total_yield



"""Test methods for compressing FASTQ"""
import logging
from pathlib import Path
from typing import Dict, List

import pytest
from pydantic.v1 import ValidationError

from cg.apps.crunchy import CrunchyAPI
from cg.apps.crunchy.files import (
    get_crunchy_metadata,
    get_fastq_to_spring_sbatch_path,
    get_log_dir,
    get_spring_archive_files,
)
from cg.apps.crunchy.models import CrunchyFile, CrunchyMetadata
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.models import CompressionData
from cg.utils import Process


def test_get_spring_metadata(spring_metadata_file: Path):
    """Test the method that fetches the SPRING metadata from a file"""
    # GIVEN a SPRING path and the path to a populated SPRING metadata file
    assert spring_metadata_file.is_file()
    assert spring_metadata_file.exists()

    # WHEN fetching the content of the file
    parsed_content: CrunchyMetadata = get_crunchy_metadata(spring_metadata_file)

    # THEN assert information about the three files is there
    assert len(parsed_content.files) == 3
    # THEN assert a list with the files is returned
    assert isinstance(parsed_content.files, list)


def test_get_spring_archive_files(crunchy_metadata_object: CrunchyMetadata):
    """Test the method that sorts the SPRING metadata into a dictionary"""
    # GIVEN a SPRING metadata content in its raw format
    assert isinstance(crunchy_metadata_object.files, list)

    # WHEN sorting the files
    sorted_content: Dict[str, CrunchyFile] = get_spring_archive_files(crunchy_metadata_object)

    # THEN assert information about the three files is there
    assert len(sorted_content) == 3
    # THEN assert a dictionary with the files is returned
    assert isinstance(sorted_content, dict)
    # THEN assert that the correct files are there
    for file_name in ["fastq_first", "fastq_second", "spring"]:
        assert file_name in sorted_content


def test_get_spring_metadata_malformed_info(
    spring_metadata_file: Path, spring_metadata: List[dict]
):
    """Test the method that fetches the SPRING metadata from a file when file is malformed"""
    # GIVEN a SPRING metadata file with missing information
    spring_metadata[0].pop("path")
    WriteFile.write_file_from_content(
        content=spring_metadata, file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # WHEN fetching the content of the file
    with pytest.raises(ValidationError):
        # THEN assert that a pydantic validation error is raised
        get_crunchy_metadata(spring_metadata_file)


def test_get_spring_metadata_wrong_number_files(
    spring_metadata_file: Path, spring_metadata: List[dict]
):
    """Test the method that fetches the SPRING metadata from a file when a file is missing"""
    # GIVEN a SPRING metadata file with missing file
    spring_metadata = spring_metadata[1:]
    WriteFile.write_file_from_content(
        content=spring_metadata, file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # WHEN fetching the content of the file
    with pytest.raises(ValidationError):
        # THEN assert that a validation error is raised
        get_crunchy_metadata(spring_metadata_file)


def test_fastq_to_spring_sbatch(
    crunchy_config: dict,
    compression_object: CompressionData,
    sbatch_process: Process,
    sbatch_job_number: int,
    caplog,
):
    """Test fastq_to_spring method"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api, and FASTQ paths

    crunchy_api = CrunchyAPI(crunchy_config)
    crunchy_api.slurm_api.process = sbatch_process
    spring_path: Path = compression_object.spring_path
    log_path: Path = get_log_dir(spring_path)
    run_name: str = compression_object.run_name
    sbatch_path: Path = get_fastq_to_spring_sbatch_path(log_dir=log_path, run_name=run_name)

    # GIVEN that the sbatch file does not exist
    assert not sbatch_path.is_file()
    # GIVEN that the pending path does not exist
    assert compression_object.pending_exists() is False

    # WHEN calling fastq_to_spring on FASTQ files
    job_number: int = crunchy_api.fastq_to_spring(compression_obj=compression_object)

    # THEN assert that the sbatch file was created
    assert sbatch_path.is_file()
    # THEN assert that correct job number was returned
    assert job_number == sbatch_job_number
    # THEN assert that the pending path was created
    assert compression_object.pending_exists() is True


def test_spring_to_fastq(
    compression_object: CompressionData,
    spring_metadata_file: Path,
    crunchy_config: dict,
    mocker,
):
    """Test SPRING to FASTQ method

    Test to decompress SPRING to FASTQ. This test will make sure that the correct sbatch content
    was submitted to the Process api
    """
    # GIVEN a crunchy-api given an existing SPRING metadata file
    assert spring_metadata_file.exists()
    mocker_submit_sbatch = mocker.patch.object(SlurmAPI, "submit_sbatch")
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the pending path does not exist
    assert compression_object.pending_exists() is False

    # WHEN calling bam_to_cram method on bam-path
    crunchy_api.spring_to_fastq(compression_obj=compression_object)

    # THEN _submit_sbatch method is called
    mocker_submit_sbatch.assert_called()
    # THEN assert that the pending path was created
    assert compression_object.pending_exists() is True



"""Parse statistics from the dragen quality metrics file"""
import csv
import logging
from pathlib import Path
from typing import Dict, Tuple

LOG = logging.getLogger(__name__)


class QualityMetrics:
    def __init__(self, quality_metrics_path: Path):
        self.quality_metrics_path = quality_metrics_path
        self.parsed_metrics = self.parse_metrics_file()

    def parse_metrics_file(
        self,
    ) -> Dict[int, dict]:
        """Parse the Dragen quality metrics file"""
        LOG.info("Parsing Dragen demultiplexing quality metrics file %s", self.quality_metrics_path)

        parsed_metrics = {}

        with open(self.quality_metrics_path, mode="r") as metrics_file:
            metrics_reader = csv.DictReader(metrics_file)
            for row in metrics_reader:
                lane = int(row["Lane"])
                read_number = row["ReadNumber"]
                sample_id = row["SampleID"]
                parsed_metrics[lane] = parsed_metrics.get(lane, {})
                row["YieldQ30"] = int(row["YieldQ30"])
                parsed_metrics[lane][(read_number, sample_id)] = row

        return self.summerize_quality_metrics(parsed_metrics=parsed_metrics)

    @staticmethod
    def summerize_quality_metrics(parsed_metrics: Dict[int, dict]) -> Dict[Tuple[str, str], dict]:
        """Summerize forward and reverse read information for each sample in each lane."""

        summarized_metrics = {}
        for lane in parsed_metrics:
            # Iterate over all samples in lane
            summarized_metrics[lane] = summarized_metrics.get(lane, {})
            for value in parsed_metrics[lane].values():
                value["YieldQ30"] = int(value["YieldQ30"])
                value["Mean Quality Score (PF)"] = float(value["Mean Quality Score (PF)"])
                value["QualityScoreSum"] = int(value["QualityScoreSum"])
                sample_id = value.get("SampleID")
                if sample_id not in summarized_metrics[lane]:
                    summarized_metrics[lane][sample_id] = value
                    continue
                summarized_metrics[lane][sample_id]["YieldQ30"] += value.get("YieldQ30")
                summarized_metrics[lane][sample_id]["Mean Quality Score (PF)"] += value.get(
                    "Mean Quality Score (PF)"
                )
                summarized_metrics[lane][sample_id]["QualityScoreSum"] += value.get(
                    "QualityScoreSum"
                )

        return summarized_metrics



"""Functions that deals with dummy samples"""
from cg.apps.demultiplex.sample_sheet.models import FlowCellSample

from typing import Type


def get_dummy_sample_name(sample_name: str) -> str:
    """Convert a string to a dummy sample name replacing spaces and parentheses with dashes."""
    return sample_name.replace(" ", "-").replace("(", "-").replace(")", "-")


def get_dummy_sample(
    flow_cell_id: str, dummy_index: str, lane: int, name: str, sample_type: Type[FlowCellSample]
) -> FlowCellSample:
    """Constructs and returns a dummy sample in Novaseq sample sheet format."""
    return sample_type(
        flowcell_id=flow_cell_id,
        lane=lane,
        sample_id=get_dummy_sample_name(sample_name=name),
        index=dummy_index,
        sample_name="indexcheck",
        project="indexcheck",
    )



from pathlib import Path
from typing import List
from cg.io.csv import read_csv, write_csv, read_csv_stream, write_csv_stream


def test_get_content_from_file(csv_file_path: Path):
    """
    Tests read CSV.
    """
    # GIVEN a csv file

    # WHEN reading the csv file
    raw_csv_content: List[List[str]] = read_csv(file_path=csv_file_path)

    # Then assert a list is returned
    assert isinstance(raw_csv_content, List)


def test_get_content_from_file_to_dict(csv_file_path: Path):
    """
    Tests read CSV into a list of dictionaries.
    """
    # GIVEN a csv file

    # WHEN reading the csv file
    raw_csv_content: List[List[str]] = read_csv(file_path=csv_file_path, read_to_dict=True)

    # Then assert a list is returned and that the first element is a dict
    assert isinstance(raw_csv_content, List)
    assert isinstance(raw_csv_content[0], dict)


def test_get_content_from_stream(csv_stream: str):
    """
    Tests read CSV stream.
    """
    # GIVEN a string in csv format

    # WHEN reading the csv content in string
    raw_content: List[List[str]] = read_csv_stream(stream=csv_stream)

    # THEN assert a list is returned
    assert isinstance(raw_content, List)

    # THEN the content should match the expected content
    expected_content = [["Lorem", "ipsum", "sit", "amet"]]
    assert raw_content == expected_content


def test_write_csv(csv_file_path: Path, csv_temp_path: Path):
    """
    Tests write CSV.
    """
    # GIVEN a csv file

    # GIVEN a file path to write to

    # WHEN reading the csv file
    raw_csv_content: List[List[str]] = read_csv(file_path=csv_file_path)

    # WHEN writing the csv file from dict
    write_csv(content=raw_csv_content, file_path=csv_temp_path)

    # THEN assert that a file was successfully created
    assert Path.exists(csv_temp_path)

    # WHEN reading it as a csv
    written_raw_csv_content: List[List[str]] = read_csv(file_path=csv_temp_path)

    # THEN assert that all data is kept
    assert raw_csv_content == written_raw_csv_content


def test_write_csv_stream(csv_stream: str):
    """
    Tests write CSV stream.
    """
    # GIVEN a list of lists

    # WHEN writing the csv stream
    written_stream: str = write_csv_stream(content=[["Lorem", "ipsum", "sit", "amet"]])

    # THEN assert that the stream is correct
    assert written_stream == csv_stream + "\n"



"""Some helper functions for working with files"""
from pathlib import Path
import os
from typing import List


def get_file_in_directory(directory: Path, file_name: str) -> Path:
    """Get a file in a directory and subdirectories.
    Raises:
        FileNotFoundError: If the directory does not exist.
    """
    if not directory.is_dir() or not directory.exists():
        raise FileNotFoundError(f"Directory {directory} does not exist")
    for directory_path, _, files in os.walk(directory):
        for file in files:
            if file_name == file:
                path_to_file = Path(directory_path, file)
                return path_to_file
    raise FileNotFoundError(f"File {file_name} not found in {directory}")


def get_files_matching_pattern(directory: Path, pattern: str) -> List[Path]:
    """Search for all files in a directory that match a pattern."""
    return list(directory.glob(pattern))



"""Test module for the Rnafusion delivery report API."""
from typing import Dict

from cg.models.report.metadata import RnafusionSampleMetadataModel
from cg.models.rnafusion.analysis import RnafusionAnalysis

from cg.meta.report.rnafusion import RnafusionReportAPI

from cg.store.models import Sample, Family


def test_get_sample_metadata(
    report_api_rnafusion: RnafusionReportAPI,
    rnafusion_sample_id: str,
    rnafusion_case_id: str,
    rnafusion_validated_metrics: Dict[str, str],
    mock_analysis_finish,
):
    """Test Rnafusion sample metadata extraction."""

    # GIVEN a Rnafusion case and associated sample
    case: Family = report_api_rnafusion.status_db.get_case_by_internal_id(
        internal_id=rnafusion_case_id
    )
    sample: Sample = report_api_rnafusion.status_db.get_sample_by_internal_id(
        internal_id=rnafusion_sample_id
    )

    # GIVEN an analysis metadata object
    latest_metadata: RnafusionAnalysis = report_api_rnafusion.analysis_api.get_latest_metadata(
        case_id=rnafusion_case_id
    )

    # WHEN getting the sample metadata
    sample_metadata: RnafusionSampleMetadataModel = report_api_rnafusion.get_sample_metadata(
        case=case, sample=sample, analysis_metadata=latest_metadata
    )

    # THEN the sample metadata should be correctly retrieved and match the expected validated metrics
    assert sample_metadata.dict() == rnafusion_validated_metrics



import logging
from datetime import datetime
from pathlib import Path
from typing import Union

from cg.models.orders.constants import OrderType
from cg.constants import (
    Pipeline,
    NA_FIELD,
    YES_FIELD,
    NO_FIELD,
    PRECISION,
    REPORT_SUPPORTED_PIPELINES,
    BALSAMIC_ANALYSIS_TYPE,
    REPORT_GENDER,
)

LOG = logging.getLogger(__name__)


def validate_empty_field(value: Union[int, str]) -> str:
    """Formats an empty value to be included in the report as N/A."""
    return str(value) if value else NA_FIELD


def validate_boolean(value: Union[bool, str]) -> str:
    """Formats a boolean value for the delivery report."""
    if isinstance(value, bool) or value:
        if str(value) == "True":
            return YES_FIELD
        if str(value) == "False":
            return NO_FIELD
    return NA_FIELD


def validate_float(value: Union[float, str]) -> str:
    """Returns a processed float value."""
    return str(round(float(value), PRECISION)) if value or isinstance(value, float) else NA_FIELD


def validate_percentage(value: Union[float, str]) -> str:
    """Returns a processed float value as a percentage."""
    return validate_float(float(value) * 100) if value else NA_FIELD


def validate_date(date: datetime) -> str:
    """Returns the date part (year, month, day) from a datetime object."""
    return str(date.date()) if date else NA_FIELD


def validate_list(value: list) -> str:
    """Formats a list elements as comma separated individual values."""
    return validate_empty_field(
        ", ".join(validate_empty_field(v) for v in value) if value else NA_FIELD
    )


def validate_path(file_path: str) -> str:
    """Returns the name of a specific file."""
    return Path(file_path).name if file_path and Path(file_path).is_file() else NA_FIELD


def validate_gender(value: str) -> str:
    """Formats the provided gender."""
    return validate_empty_field(REPORT_GENDER.get(value))


def validate_rml_sample(prep_category: str) -> str:
    """Checks if a specific sample is a RML one."""
    if prep_category == OrderType.RML:
        LOG.error("The delivery report generation does not support RML samples")
        raise ValueError
    return validate_empty_field(prep_category)


def validate_balsamic_analysis_type(value: str) -> str:
    """Translates the BALSAMIC analysis type string to an accepted value for the delivery report."""
    return validate_empty_field(BALSAMIC_ANALYSIS_TYPE.get(value))


def validate_supported_pipeline(cls, values: dict) -> dict:
    """Validates if the report generation supports a specific pipeline and analysis type."""
    if values and values.get("pipeline") and values.get("customer_pipeline"):
        # Checks that the requested analysis and the executed one match
        if values.get("pipeline") != values.get("customer_pipeline"):
            LOG.error(
                f"The analysis requested by the customer ({values.get('customer_pipeline')}) does not match the one "
                f"executed ({values.get('pipeline')})"
            )
            raise ValueError
        # Check that the generation of the report supports the data analysis executed on the case
        if values.get("pipeline") not in REPORT_SUPPORTED_PIPELINES:
            LOG.error(
                f"The pipeline {values.get('pipeline')} does not support delivery report generation"
            )
            raise ValueError
    # Validates the analysis type
    if Pipeline.BALSAMIC in values.get("pipeline"):
        values["type"] = validate_balsamic_analysis_type(values["type"])
    return values



from typing import Dict
import pytest

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import (
    ConversionResult,
    DemuxResult,
    IndexMetric,
    ReadMetric,
)


@pytest.fixture
def valid_bcl2fastq_metrics_data() -> Dict:
    """Example structure of a valid stats.json file containing sequecing metrics generated by bcl2fastq."""
    return {
        "Flowcell": "AB1",
        "RunNumber": 1,
        "RunId": "RUN1",
        "ReadInfosForLanes": [
            {
                "LaneNumber": 1,
            }
        ],
        "ConversionResults": [
            {
                "LaneNumber": 1,
                "Yield": 1000,
                "DemuxResults": [
                    {
                        "SampleId": "S1",
                        "SampleName": "Sample1",
                        "IndexMetrics": [
                            {
                                "IndexSequence": "ATGC",
                                "MismatchCounts": {"0": 24470341, "1": 5093729},
                            }
                        ],
                        "NumberReads": 1,
                        "Yield": 100,
                        "ReadMetrics": [
                            {"ReadNumber": 1, "Yield": 100, "YieldQ30": 90, "QualityScoreSum": 100}
                        ],
                    }
                ],
            }
        ],
    }


@pytest.fixture
def conversion_result() -> ConversionResult:
    return ConversionResult(
        LaneNumber=1,
        Yield=1000,
        DemuxResults=[
            DemuxResult(
                SampleId="S1",
                SampleName="Sample1",
                IndexMetrics=[
                    IndexMetric(IndexSequence="ATGC", MismatchCounts={"0": 24470341, "1": 5093729})
                ],
                NumberReads=1,
                Yield=100,
                ReadMetrics=[ReadMetric(ReadNumber=1, Yield=100, YieldQ30=90, QualityScoreSum=100)],
            )
        ],
    )



import collections

from pydantic.v1 import BaseModel, validator

from cg.constants.nextflow import DELIVER_FILE_HEADERS


def replace_dict_values(replace_map: dict, my_dict: dict) -> dict:
    for str_to_replace, with_value in replace_map.items():
        for key, value in my_dict.items():
            if not value:
                value = "~"
            my_dict.update({key: value.replace(str_to_replace, with_value)})
    return my_dict


class NextflowDeliverables(BaseModel):
    """Nextflow deliverables model

    Attributes:
        deliverables: dictionary containing format, path, path_index, step, tag and id keys
    """

    deliverables: dict

    @validator("deliverables")
    def headers(cls, v: dict) -> None:
        """Validate header format."""
        if collections.Counter(list(v.keys())) != collections.Counter(DELIVER_FILE_HEADERS):
            raise ValueError(
                f"Headers are not matching the standard header format: {DELIVER_FILE_HEADERS}"
            )
        for key, value in v.items():
            if not value and key != "path_index":
                raise ValueError("An entry other than path_index is empty!")
